# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import grp
import os
import pwd

import iocage.lib.helpers


class Storage:
    def __init__(self, jail,
                 zfs=None,
                 safe_mode=True,
                 logger=None):

        self.logger = iocage.lib.helpers.init_logger(self, logger)
        self.zfs = iocage.lib.helpers.init_zfs(self, zfs)
        self.jail = jail

        # safe-mody only attaches zfs datasets to jails that were tagged with
        # jailed=on already exist
        self.safe_mode = safe_mode

    def clone_release(self, release):
        self.clone_zfs_dataset(
            release.root_dataset_name,
            self.jail.root_dataset_name
        )
        jail_name = self.jail.humanreadable_name
        self.logger.verbose(
            f"Cloned release '{release.name}' to {jail_name}",
            jail=self.jail
        )

    def delete_dataset_recursive(self, dataset, delete_snapshots=True):

        for child in dataset.children:
            self.delete_dataset_recursive(child)

        if dataset.mountpoint is not None:
            self.logger.spam("Unmounting {dataset.name}")
            dataset.umount()

        origin = None
        if delete_snapshots is False:
            origin_property = dataset.properties["origin"]
            if origin_property.value != "":
                origin = origin_property

        self.logger.verbose("Deleting dataset {dataset.name}")
        dataset.delete()

        if origin is not None:
            self.logger.verbose("Deleting snapshot {origin}")
            origin_snapshot = self.zfs.get_snapshot(origin)
            origin_snapshot.delete()

    def clone_zfs_dataset(self, source, target):

        snapshot_name = f"{source}@{self.jail.name}"

        # delete target dataset if it already exists
        try:
            existing_dataset = self.zfs.get_dataset(target)
            self.logger.verbose(
                f"Deleting existing dataset {target}",
                jail=self.jail
            )
            if existing_dataset.mountpoint is not None:
                existing_dataset.umount()
            existing_dataset.delete()
            del existing_dataset
        except:
            pass

        # delete existing snapshot if existing
        existing_snapshot = None
        try:
            existing_snapshot = self.zfs.get_snapshot(snapshot_name)
        except:
            pass

        if existing_snapshot:
            self.logger.verbose(
                f"Deleting existing snapshot {snapshot_name}",
                jail=self.jail
            )
            existing_snapshot.delete()

        # snapshot release
        self.zfs.get_dataset(source).snapshot(snapshot_name)
        snapshot = self.zfs.get_snapshot(snapshot_name)

        # clone snapshot
        try:
            self.logger.verbose(
                f"Cloning snapshot {snapshot_name} to {target}",
                jail=self.jail
            )
            snapshot.clone(target)
        except:
            parent = "/".join(target.split("/")[:-1])
            self.logger.debug(
                "Cloning was unsuccessful - "
                f"trying to create the parent dataset '{parent}' first",
                jail=self.jail
            )
            self.zfs.create_dataset(parent)
            snapshot.clone(target)

        target_dataset = self.zfs.get_dataset(target)
        target_dataset.mount()
        self.logger.verbose(
            f"Successfully cloned {source} to {target}",
            jail=self.jail
        )

    def create_jail_mountpoint(self, basedir: str):
        basedir = f"{self.jail.root_dataset.mountpoint}/{basedir}"
        if not os.path.isdir(basedir):
            self.logger.verbose(f"Creating mountpoint {basedir}")
            os.makedirs(basedir)

    def _mount_procfs(self):
        try:
            if self.jail.config["mount_procfs"]:
                iocage.lib.helpers.exec([
                    "mount"
                    "-t",
                    "procfs"
                    "proc"
                    f"{self.jail.root_dataset.mountpoint}/proc"
                ])
        except:
            raise iocage.lib.errors.MountFailed("procfs")

    # ToDo: Remove unused function?
    def _mount_linprocfs(self):
        try:
            if not self.jail.config["mount_linprocfs"]:
                return
        except:
            pass

        linproc_path = self._jail_mkdirp("/compat/linux/proc")

        try:
            if self.jail.config["mount_procfs"]:
                iocage.lib.helpers.exec([
                    "mount"
                    "-t",
                    "linprocfs",
                    "linproc",
                    linproc_path
                ])
        except:
            raise iocage.lib.errors.MountFailed("linprocfs")

    def _unmount_local(self, dataset):
        if dataset.mountpoint:
            dataset.unmount()

    def _jail_mkdirp(
        self,
        directory: str,
        permissions=0o775,
        user: str="root",
        group: str="wheel"
    ) -> str:

        uid = pwd.getpwnam(user).pw_uid
        gid = grp.getgrnam(group).gr_gid
        folder = f"{self.jail.root_dataset.mountpoint}{directory}"
        if not os.path.isdir(folder):
            os.makedirs(folder, permissions)
            os.chown(folder, uid, gid, follow_symlinks=False)  # type: ignore
        return os.path.abspath(folder)



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import iocage.lib.helpers


class NetworkInterface:
    ifconfig_command = "/sbin/ifconfig"
    dhclient_command = "/sbin/dhclient"

    def __init__(self,
                 name="vnet0",
                 ipv4_addresses=[],
                 ipv6_addresses=[],
                 mac=None,
                 mtu=None,
                 description=None,
                 rename=None,
                 addm=None,
                 vnet=None,
                 jail=None,
                 extra_settings=[],
                 auto_apply=True,
                 logger=None):

        self.jail = jail

        self.name = name
        self.ipv4_addresses = ipv4_addresses
        self.ipv6_addresses = ipv6_addresses

        self.extra_settings = extra_settings
        self.settings = {}

        if mac:
            self.settings["link"] = mac

        if mtu:
            self.settings["mtu"] = str(mtu)

        if description:
            self.settings["description"] = f"\"{description}\""

        if vnet:
            self.settings["vnet"] = vnet

        if addm:
            self.settings["addm"] = addm

        # rename interface when applying settings next time
        if isinstance(rename, str):
            self.rename = True
            self.settings["name"] = rename
        else:
            self.rename = False

        if auto_apply:
            self.apply()

    def apply(self):
        self.apply_settings()
        self.apply_addresses()

    def apply_settings(self):
        command = [self.ifconfig_command, self.name]
        for key in self.settings:
            command.append(key)
            command.append(self.settings[key])

        if self.extra_settings:
            command += self.extra_settings

        self.exec(command)

        # update name when the interface was renamed
        if self.rename:
            self.name = self.settings["name"]
            del self.settings["name"]
            self.rename = False

    def apply_addresses(self):
        self.__apply_addresses(self.ipv4_addresses, ipv6=False)
        self.__apply_addresses(self.ipv6_addresses, ipv6=True)

    def __apply_addresses(self, addresses, ipv6=False):
        family = "inet6" if ipv6 else "inet"
        for address in addresses:
            if address.lower() == "dhcp":
                command = [self.dhclient_command, self.name]
            else:
                command = [self.ifconfig_command, self.name, family, address]
            self.exec(command)

    def exec(self, command, force_local=False):
        if self.__is_jail():
            return self.jail.exec(command)
        else:
            return iocage.lib.helpers.exec(command)

    def __is_jail(self):
        return self.jail is not None



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
"""fetch module for the cli."""
import click

import iocage.lib.Host
import iocage.lib.Prompts
import iocage.lib.Release
import iocage.lib.errors


__rootcmd__ = True


@click.command(context_settings=dict(
    max_content_width=400, ),
    name="fetch", help="Fetch a version of FreeBSD for jail usage or a"
                       " preconfigured plugin.")
@click.pass_context
@click.option("--url", "-u",
              help="Remote URL with path to the release/snapshot directory")
@click.option("--file", "-F", multiple=True,
              help="Specify the files to fetch from the mirror.")
@click.option("--release", "-r",
              # type=release_choice(),
              help="The FreeBSD release to fetch.")
@click.option("--update/--no-update", "-U/-NU", default=True,
              help="Update the release to the latest patch level.")
@click.option("--fetch-updates/--no-fetch-updates", default=True,
              help="Skip fetching release updates")
# Compat
@click.option("--http", "-h", default=False,
              help="Have --server define a HTTP server instead.", is_flag=True)
# Compat files
@click.option("--files", multiple=True,
              help="Specify the files to fetch from the mirror. "
                   "(Deprecared: renamed to --file)")
def cli(ctx, **kwargs):
    logger = ctx.parent.logger
    host = iocage.lib.Host.HostGenerator(logger=logger)
    prompts = iocage.lib.Prompts.Prompts(host=host, logger=logger)

    release_input = kwargs["release"]
    if release_input is None:
        try:
            release = prompts.release()
        except iocage.lib.errors.DefaultReleaseNotFound:
            exit(1)
    else:
        try:
            release = iocage.lib.Release.ReleaseGenerator(
                name=release_input,
                host=host,
                logger=logger
            )
        except:
            logger.error(f"Invalid Release '{release_input}'")
            exit(1)

    url_or_files_selected = False

    if is_option_enabled(kwargs, "url"):
        release.mirror_url = kwargs["url"]
        url_or_files_selected = True

    if is_option_enabled(kwargs, "files"):
        release.assets = list(kwargs["files"])
        url_or_files_selected = True

    if (url_or_files_selected is False) and (release.available is False):
        logger.error(f"The release '{release.name}' is not available")
        exit(1)

    fetch_updates = bool(kwargs["fetch_updates"])
    ctx.parent.print_events(release.fetch(
        update=kwargs["update"],
        fetch_updates=fetch_updates
    ))

    exit(0)


def is_option_enabled(args, name):
    try:
        value = args[name]
        if value:
            return True
    except KeyError:
        pass

    return False



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
from iocage.lib import errors, events
from iocage.lib.Host import Host
from iocage.lib.Jail import Jail
from iocage.lib.Jails import Jails
from iocage.lib.Logger import Logger
from iocage.lib.Release import Release
from iocage.lib.Releases import Releases



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import typing
import os
import subprocess
import uuid
import shlex

import iocage.lib.Types
import iocage.lib.errors
import iocage.lib.events
import iocage.lib.helpers
import iocage.lib.DevfsRules
import iocage.lib.Host
import iocage.lib.Config.Jail.JailConfig
import iocage.lib.Network
import iocage.lib.NullFSBasejailStorage
import iocage.lib.Release
import iocage.lib.StandaloneJailStorage
import iocage.lib.Storage
import iocage.lib.ZFSBasejailStorage
import iocage.lib.ZFSShareStorage
import iocage.lib.LaunchableResource
import iocage.lib.Config.Jail.File.Fstab


class JailResource(iocage.lib.LaunchableResource.LaunchableResource):

    _jail: 'JailGenerator' = None
    _fstab: 'iocage.lib.Config.Jail.File.Fstab.Fstab' = None

    def __init__(
        self,
        host: 'iocage.lib.Host.HostGenerator',
        jail: 'JailGenerator'=None,
        **kwargs
    ) -> None:

        self.__jails_dataset_name = host.datasets.jails.name
        self.host = iocage.lib.helpers.init_host(self, host)

        self._jail = jail

        iocage.lib.LaunchableResource.LaunchableResource.__init__(
            self,
            **kwargs
        )

    @property
    def jail(self) -> 'JailGenerator':
        """
        Jail instance that belongs to the resource

        Usually the resource becomes inherited from the jail itself.
        It can still be used linked to a foreign jail by passing jail as
        named attribute to the __init__ function
        """
        if self._jail is not None:
            return self._jail

        # is instance of Jail itself
        if isinstance(self, JailGenerator):
            return self

        raise Exception("This resource is not a jail or not linked to one")

    @property
    def fstab(self) -> 'iocage.lib.Config.Jail.File.Fstab.Fstab':
        if self._fstab is None:
            self._fstab = iocage.lib.Config.Jail.File.Fstab.Fstab(
                jail=self.jail,
                release=self.jail.release,
                logger=self.logger,
                host=self.jail.host
            )
        return self._fstab

    @property
    def dataset_name(self) -> str:
        """
        Name of the jail base ZFS dataset

        If the resource has no dataset or dataset_name assigned yet,
        the jail id is used to find name the dataset
        """
        try:
            return self._assigned_dataset_name
        except:
            pass

        return self._dataset_name_from_jail_name

    @dataset_name.setter
    def dataset_name(self, value: str) -> None:
        self._dataset_name = value

    @property
    def _dataset_name_from_jail_name(self) -> str:

        jail_id = self.jail.config["id"]
        if jail_id is None:
            raise iocage.lib.errors.JailUnknownIdentifier()

        return f"{self.__jails_dataset_name}/{jail_id}"

    def get(self, key: str) -> typing.Any:
        try:
            out = self.jail.config[key]
            return out
        except KeyError:
            pass

        return iocage.lib.Resource.Resource.get(self, key)


class JailGenerator(JailResource):
    """
    iocage unit orchestrates a jail's configuration and manages state

    Jails are represented as a zfs dataset `zpool/iocage/jails/<NAME>`

    Directory Structure:

        zpool/iocage/jails/<NAME>: The jail's dataset containing it's
            configuration and root dataset. iocage-legacy used to store
            a jails configuration as ZFS properties on this dataset. Even
            though the modern JSON config mechanism is preferred.

        zpool/iocage/jails/<NAME>/root: This directory is the dataset
            used as jail's root when starting a jail. Usually the clone source
            of a root dataset is a snapshot of the release's root dataset.

        zpool/iocage/jails/<NAME>/config.json: Jails configured with the latest
            configuration style store their information in a JSON file. When
            this file is found in the jail's dataset, libiocage assumes
            the jail to be a JSON-style jail and ignores other configuration
            mechanisms

        zpool/iocage/jails/<NAME>/config: Another compatible configuration
            mechanism is a UCL file. It's content is only taken into account if
            no JSON or ZFS configuration was found

    Jail Types:

        Standalone: The /root dataset gets cloned from a release at creation
            time. It it not affected by changes to the Release and persists all
            data within the jail

        NullFS Basejail: The fastest method to spawn a basejail by mounting
            read-only directories from the release's root dataset by creating
            a snapshot of the release on each boot of the jail. When a release
            is updated, the jail is updated as well on the next reboot. This
            type is the one used by the Python implementation of iocage.

        ZFS Basejail: Legacy basejails used to clone individual datasets from a
            release (stored in `zpool/iocage/base/<RELEASE>)
    """

    _class_storage = iocage.lib.Storage.Storage

    jail_state: dict = None

    def __init__(
        self,
        data: typing.Union[str, typing.Dict[str, typing.Any]]={},
        zfs=None,
        host=None,
        logger=None,
        new=False,
        **resource_args
    ) -> None:
        """
        Initializes a Jail

        Args:

            data (string|dict):
                Jail configuration dict or jail name as string identifier.

            zfs (libzfs.ZFS): (optional)
                Inherit an existing libzfs.ZFS() instance from ancestor classes

            host (iocage.lib.Host): (optional)
                Inherit an existing Host instance from ancestor classes

            logger (iocage.lib.Logger): (optional)
                Inherit an existing Logger instance from ancestor classes

        """

        self.logger = iocage.lib.helpers.init_logger(self, logger)
        self.zfs = iocage.lib.helpers.init_zfs(self, zfs)
        self.host = iocage.lib.helpers.init_host(self, host)

        if isinstance(data, str):
            data = {
                "id": self._resolve_name(data)
            }

        self.config = iocage.lib.Config.Jail.JailConfig.JailConfig(
            data=data,
            host=self.host,
            jail=self,
            logger=self.logger
        )

        self.storage = self._class_storage(
            safe_mode=False,
            jail=self,
            logger=self.logger,
            zfs=self.zfs
        )

        JailResource.__init__(
            self,
            jail=self,
            host=self.host,
            logger=self.logger,
            zfs=self.zfs,
            **resource_args
        )

        if new is False:
            self.config.read(data=self.read_config())
            if self.config["id"] is None:
                self.config["id"] = self.dataset_name.split("/").pop()

    def start(
        self,
        quick: bool=False
    ) -> typing.Generator['iocage.lib.events.IocageEvent', None, None]:
        """
        Start the jail.
        """

        self.require_jail_existing()
        self.require_jail_stopped()

        release = self.release

        events: typing.Any = iocage.lib.events
        jailLaunchEvent = events.JailLaunch(jail=self)
        jailVnetConfigurationEvent = events.JailVnetConfiguration(jail=self)
        JailZfsShareMount = events.JailZfsShareMount(jail=self)
        jailServicesStartEvent = events.JailServicesStart(jail=self)

        yield jailLaunchEvent.begin()

        if self.basejail_backend is not None:
            self.basejail_backend.apply(self.storage, release)

        if quick is False:
            self._save_autoconfig()

        self._run_hook("prestart")
        self._launch_jail()

        yield jailLaunchEvent.end()

        if self.config["vnet"]:
            yield jailVnetConfigurationEvent.begin()
            self._start_vimage_network()
            self._configure_routes()
            yield jailVnetConfigurationEvent.end()

        self._limit_resources()
        self._configure_nameserver()

        if self.config["jail_zfs"] is True:
            yield JailZfsShareMount.begin()
            iocage.lib.ZFSShareStorage.ZFSShareStorage.mount_zfs_shares(
                self.storage
            )
            yield JailZfsShareMount.end()

        if self.config["exec_start"] is not None:
            yield jailServicesStartEvent.begin()
            self._start_services()
            yield jailServicesStartEvent.end()

        self._run_hook("poststart")

    @property
    def basejail_backend(self):

        if self.config["basejail"] is False:
            return None

        if self.config["basejail_type"] == "nullfs":
            return iocage.lib.NullFSBasejailStorage.NullFSBasejailStorage

        if self.config["basejail_type"] == "zfs":
            return iocage.lib.ZFSBasejailStorage.ZFSBasejailStorage

        return None

    def _run_hook(self, hook_name: str):

        key = f"exec_{hook_name}"
        value = self.config[key]

        if value == "/usr/bin/true":
            return

        self.logger.verbose(
            f"Running {hook_name} hook for {self.humanreadable_name}"
        )

        lex = shlex.shlex(value)
        lex.whitespace_split = True
        command = list(lex)  # type: ignore

        return iocage.lib.helpers.exec(
            command,
            logger=self.logger,
            env=self.env
        )

    def _start_services(self):
        command = self.config["exec_start"].strip().split()
        self.logger.debug(f"Running exec_start on {self.humanreadable_name}")
        self.exec(command)

    def stop(
        self,
        force: bool=False
    ) -> typing.Generator['iocage.lib.events.IocageEvent', None, None]:
        """
        Stop a jail.

        Args:

            force (bool): (default=False)
                Ignores failures and enforces teardown if True
        """

        if force is True:
            for event in self._force_stop():
                yield event
            return

        self.require_jail_existing()
        self.require_jail_running()

        events: typing.Any = iocage.lib.events
        jailDestroyEvent = events.JailDestroy(self)
        jailNetworkTeardownEvent = events.JailNetworkTeardown(self)
        jailMountTeardownEvent = events.JailMountTeardown(self)

        self._run_hook("prestop")

        yield jailDestroyEvent.begin()
        self._destroy_jail()
        yield jailDestroyEvent.end()

        if self.config["vnet"]:
            yield jailNetworkTeardownEvent.begin()
            self._stop_vimage_network()
            yield jailNetworkTeardownEvent.end()

        yield jailMountTeardownEvent.begin()
        self._teardown_mounts()
        yield jailMountTeardownEvent.end()

        self.update_jail_state()

    def destroy(self, force=False):
        """
        Destroy a Jail and it's datasets

        Args:

            force (bool): (default=False)
                This flag enables whether an existing jail should be shut down
                before destroying the dataset. By default destroying a jail
                requires it to be stopped.
        """

        self.update_jail_state()

        if self.running is True and force is True:
            self.stop(force=True)
        else:
            self.require_jail_stopped()

        self.storage.delete_dataset_recursive(self.dataset)

    def rename(self, new_name: str):
        """
        Change the name of a jail
        """

        self.require_jail_existing()
        self.require_jail_stopped()

        current_id = self.config["id"]
        dataset = self.dataset
        current_dataset_name = dataset.name

        self.config["id"] = new_name  # validates new_name
        self.logger.debug(f"Renaming jail {current_id} to {new_name}")
        try:
            new_dataset_name = self._dataset_name_from_jail_name
            dataset.rename(new_dataset_name)
            self.logger.verbose(
                f"Dataset {current_dataset_name} renamed to {new_dataset_name}"
            )
        except:
            self.config["id"] = current_id
            raise

    def _force_stop(self):

        successful = True

        events: typing.Any = iocage.lib.events
        jailDestroyEvent = events.JailDestroy(self)
        jailNetworkTeardownEvent = events.JailNetworkTeardown(self)
        jailMountTeardownEvent = events.JailMountTeardown(self)

        try:
            self._run_hook("prestop")
        except:
            successful = False
            self.logger.warn("pre-stop script failed")

        yield jailDestroyEvent.begin()
        try:
            self._destroy_jail()
            self.logger.debug(f"{self.humanreadable_name}: jail destroyed")
            yield jailDestroyEvent.end()
        except Exception as e:
            successful = False
            yield jailDestroyEvent.skip()

        if self.config["vnet"]:
            yield jailNetworkTeardownEvent.begin()
            try:
                self._stop_vimage_network()
                self.logger.debug(f"{self.humanreadable_name}: VNET stopped")
                yield jailNetworkTeardownEvent.end()
            except Exception as e:
                successful = False
                yield jailNetworkTeardownEvent.skip()

        yield jailMountTeardownEvent.begin()
        try:
            self._teardown_mounts()
            self.logger.debug(f"{self.humanreadable_name}: mounts destroyed")
            yield jailMountTeardownEvent.end()
        except Exception as e:
            successful = False
            yield jailMountTeardownEvent.skip()

        try:
            self.update_jail_state()
        except Exception as e:
            successful = False
            self.logger.warn(str(e))

        return successful

    def create(self, release_name):
        """
        Create a Jail from a Release

        Args:

            release_name (string):
                The jail is created from the release matching the name provided
        """

        if self.config["id"] is None:
            self.config["id"] = str(uuid.uuid4())

        self.require_jail_not_existing()

        # check if release exists
        release = iocage.lib.Release.Release(
            name=release_name,
            host=self.host,
            zfs=self.zfs,
            logger=self.logger
        )

        if release.fetched is False:
            raise iocage.lib.errors.ReleaseNotFetched(
                name=release_name,
                logger=self.logger
            )

        self.config["release"] = release.name

        self.logger.verbose(
            f"Creating jail '{self.config['id']}'",
            jail=self
        )

        for key, value in self.config.data.items():
            msg = f"{key} = {value}"
            self.logger.spam(msg, jail=self, indent=1)

        self.create_resource()
        self.get_or_create_dataset("root")
        self._update_fstab()

        backend = None

        is_basejail = self.config.get("basejail", False)
        if not is_basejail:
            backend = iocage.lib.StandaloneJailStorage.StandaloneJailStorage
        if is_basejail and self.config["basejail_type"] == "nullfs":
            backend = iocage.lib.NullFSBasejailStorage.NullFSBasejailStorage
        elif is_basejail and self.config["basejail_type"] == "zfs":
            backend = iocage.lib.ZFSBasejailStorage.ZFSBasejailStorage

        if backend is not None:
            backend.setup(self.storage, release)

        self.config.data["release"] = release.name
        self.save()

    def save(self):
        self.write_config(self.config.data)
        self._save_autoconfig()

    def _save_autoconfig(self):
        """
        Saves auto-generated files
        """
        self.rc_conf.save()
        self._update_fstab()

    def _update_fstab(self) -> None:

        if self.config["basejail_type"] == "nullfs":
            self.fstab.release = self.release
        else:
            self.fstab.release = None

        self.fstab.update_and_save()

    def exec(self, command, **kwargs):
        """
        Execute a command in a started jail

        command (list):
            A list of command and it's arguments

            Example: ["/usr/bin/whoami"]
        """

        command = ["/usr/sbin/jexec", self.identifier] + command

        return iocage.lib.helpers.exec(
            command,
            logger=self.logger,
            env=self.env,
            **kwargs
        )

    def passthru(self, command):
        """
        Execute a command in a started jail ans passthrough STDIN and STDOUT

        command (list):
            A list of command and it's arguments

            Example: ["/bin/sh"]
        """

        if isinstance(command, str):
            command = [command]

        return iocage.lib.helpers.exec_passthru(
            [
                "/usr/sbin/jexec",
                self.identifier
            ] + command,
            logger=self.logger
        )

    def exec_console(self):
        """
        Shortcut to drop into a shell of a started jail
        """
        return self.passthru(
            ["/usr/bin/login"] + self.config["login_flags"]
        )

    def _destroy_jail(self):

        command = ["jail", "-r"]
        command.append(self.identifier)

        subprocess.check_output(
            command,
            shell=False,
            stderr=subprocess.DEVNULL
        )

    @property
    def _dhcp_enabled(self):
        """
        True if any ip4_addr uses DHCP
        """
        if self.config["ip4_addr"] is None:
            return False

        return ("dhcp" in self.config["ip4_addr"].networks)

    @property
    def devfs_ruleset(self):
        """
        The number of the jails devfs ruleset

        When a new combination of the base ruleset specified in
        jail.config["devfs_ruleset"] and rules automatically added by iocage
        appears, the according rule is automatically created and added to the
        /etc/devfs.rules file on the host
        """

        # users may reference a rule by numeric identifier or name
        # numbers are automatically selected, so it's advisable to use names
        try:
            configured_devfs_ruleset = self.host.devfs.find_by_number(
                int(self.config["devfs_ruleset"])
            )
        except ValueError:
            configured_devfs_ruleset = self.host.devfs.find_by_name(
                self.config["devfs_ruleset"]
            )

        devfs_ruleset = iocage.lib.DevfsRules.DevfsRuleset()
        devfs_ruleset.clone(configured_devfs_ruleset)

        if self._dhcp_enabled:
            devfs_ruleset.append("add path 'bpf*' unhide")

        # create if the final rule combination does not exist as ruleset
        if devfs_ruleset not in self.host.devfs:
            self.logger.verbose("New devfs ruleset combination")
            # note: name and number of devfs_ruleset are both None
            new_ruleset_number = self.host.devfs.new_ruleset(devfs_ruleset)
            self.host.devfs.save()
            return new_ruleset_number
        else:
            ruleset_line_position = self.host.devfs.index(devfs_ruleset)
            return self.host.devfs[ruleset_line_position].number

    def _launch_jail(self):

        command = ["jail", "-c"]

        if self.config["vnet"]:
            command.append('vnet')
        else:

            if self.config["ip4_addr"] is not None:
                ip4_addr = self.config["ip4_addr"]
                command += [
                    f"ip4.addr={ip4_addr}",
                    f"ip4.saddrsel={self.config['ip4_saddrsel']}",
                    f"ip4={self.config['ip4']}",
                ]

            if self.config['ip6_addr'] is not None:
                ip6_addr = self.config['ip6_addr']
                command += [
                    f"ip6.addr={ip6_addr}",
                    f"ip6.saddrsel={self.config['ip6_saddrsel']}",
                    f"ip6={self.config['ip6']}",
                ]

        command += [
            f"name={self.identifier}",
            f"host.hostname={self.config['host_hostname']}",
            f"host.domainname={self.config['host_domainname']}",
            f"path={self.root_dataset.mountpoint}",
            f"securelevel={self.config['securelevel']}",
            f"host.hostuuid={self.name}",
            f"devfs_ruleset={self.devfs_ruleset}",
            f"enforce_statfs={self.config['enforce_statfs']}",
            f"children.max={self.config['children_max']}",
            f"allow.set_hostname={self.config['allow_set_hostname']}",
            f"allow.sysvipc={self.config['allow_sysvipc']}"
        ]

        if self.host.userland_version > 10.3:
            command += [
                f"sysvmsg={self.config['sysvmsg']}",
                f"sysvsem={self.config['sysvsem']}",
                f"sysvshm={self.config['sysvshm']}"
            ]

        command += [
            f"allow.raw_sockets={self.config['allow_raw_sockets']}",
            f"allow.chflags={self.config['allow_chflags']}",
            f"allow.mount={self.config['allow_mount']}",
            f"allow.mount.devfs={self.config['allow_mount_devfs']}",
            f"allow.mount.nullfs={self.config['allow_mount_nullfs']}",
            f"allow.mount.procfs={self.config['allow_mount_procfs']}",
            f"allow.mount.zfs={self.config['allow_mount_zfs']}",
            f"allow.quotas={self.config['allow_quotas']}",
            f"allow.socket_af={self.config['allow_socket_af']}",
            f"exec.stop={self.config['exec_stop']}",
            f"exec.clean={self.config['exec_clean']}",
            f"exec.timeout={self.config['exec_timeout']}",
            f"stop.timeout={self.config['stop_timeout']}",
            f"mount.fstab={self.fstab.path}",
            f"mount.devfs={self.config['mount_devfs']}"
        ]

        if self.host.userland_version > 9.3:
            command += [
                f"mount.fdescfs={self.config['mount_fdescfs']}",
                f"allow.mount.tmpfs={self.config['allow_mount_tmpfs']}"
            ]

        command += [
            "allow.dying",
            f"exec.consolelog={self.logfile_path}",
            "persist"
        ]

        humanreadable_name = self.humanreadable_name
        try:
            iocage.lib.helpers.exec(command, logger=self.logger)
            self.update_jail_state()
            self.logger.verbose(
                f"Jail '{humanreadable_name}' started with JID {self.jid}",
                jail=self
            )
        except subprocess.CalledProcessError as exc:
            code = exc.returncode
            self.logger.error(
                f"Jail '{humanreadable_name}' failed with exit code {code}",
                jail=self
            )
            raise

    @property
    def networks(self) -> list:

        networks = []

        nics = self.config["interfaces"]

        if nics is None:
            return []

        for nic in nics:

            bridges = list(self.config["interfaces"][nic])

            try:
                ipv4_addresses = self.config["ip4_addr"][nic]
            except:
                ipv4_addresses = []

            try:
                ipv6_addresses = self.config["ip6_addr"][nic]
            except:
                ipv6_addresses = []

            net = iocage.lib.Network.Network(
                jail=self,
                nic=nic,
                ipv4_addresses=ipv4_addresses,
                ipv6_addresses=ipv6_addresses,
                bridges=bridges,
                logger=self.logger
            )
            networks.append(net)

        return networks

    def _start_vimage_network(self) -> None:
        self.logger.debug("Starting VNET/VIMAGE", jail=self)
        for network in self.networks:
            network.setup()

    def _stop_vimage_network(self) -> None:
        for network in self.networks:
            network.teardown()

    def _configure_nameserver(self) -> None:
        self.config["resolver"].apply(self)

    def _limit_resources(self) -> None:

        for key, limit, action in map(
            lambda name: (name, ) + self._get_resource_limit(name),
            self._resource_limit_config_keys
        ):

            if (limit is None) and (action is None):
                # this resource is not limited (limit disabled)
                continue

            command = [
                "/usr/bin/rctl",
                "-a",
                f"jail:{self.identifier}:{key}:{action}={limit}"
            ]
            iocage.lib.helpers.exec(command, logger=self.logger)

    @property
    def _resource_limit_config_keys(self):
        return [
            "cputime",
            "datasize",
            "stacksize",
            "coredumpsize",
            "memoryuse",
            "memorylocked",
            "maxproc",
            "openfiles",
            "vmemoryuse",
            "pseudoterminals",
            "swapuse",
            "nthr",
            "msgqqueued",
            "msgqsize",
            "nmsgq",
            "nsem",
            "nsemop",
            "nshm",
            "shmsize",
            "wallclock",
            "pcpu",
            "readbps",
            "writebps",
            "readiops",
            "writeiops"
        ]

    def _get_resource_limit(self, key: str) -> typing.Tuple[str, str]:
        try:
            if isinstance(self.config[key], str):
                return self._parse_resource_limit(self.config[key])
        except:
            pass

        return None, None

    def _parse_resource_limit(
        self,
        value: str
    ) -> typing.Tuple[str, str]:

        limit, action = value.split(":", maxsplit=1)
        return limit, action

    def _configure_routes(self) -> None:

        defaultrouter = self.config["defaultrouter"]
        defaultrouter6 = self.config["defaultrouter6"]

        if not defaultrouter or defaultrouter6:
            self.logger.spam("no static routes configured")
            return

        if defaultrouter:
            self._configure_route(defaultrouter)

        if defaultrouter6:
            self._configure_route(defaultrouter6, ipv6=True)

    def _configure_route(self, gateway, ipv6=False) -> None:

        ip_version = 4 + 2 * (ipv6 is True)

        self.logger.verbose(
            f"setting default IPv{ip_version} gateway to {gateway}",
            jail=self
        )

        command = ["/sbin/route", "add"] + \
            (["-6"] if (ipv6 is True) else []) + ["default", gateway]

        self.exec(command)

    def require_jail_not_existing(self, **kwargs) -> None:
        """
        Raise JailAlreadyExists exception if the jail already exists
        """
        if self.exists:
            raise iocage.lib.errors.JailAlreadyExists(
                jail=self,
                logger=self.logger,
                **kwargs
            )

    def require_jail_existing(self, **kwargs) -> None:
        """
        Raise JailDoesNotExist exception if the jail does not exist
        """
        if not self.exists:
            raise iocage.lib.errors.JailDoesNotExist(
                jail=self,
                logger=self.logger,
                **kwargs
            )

    def require_jail_stopped(self, **kwargs) -> None:
        """
        Raise JailAlreadyRunning exception if the jail is runninhg
        """
        if self.running:
            raise iocage.lib.errors.JailAlreadyRunning(
                jail=self,
                logger=self.logger,
                **kwargs
            )

    def require_jail_running(self, **kwargs) -> None:
        """
        Raise JailNotRunning exception if the jail is stopped
        """
        if not self.running:
            raise iocage.lib.errors.JailNotRunning(
                jail=self,
                logger=self.logger,
                **kwargs
            )

    def update_jail_state(self) -> None:
        """
        Invoke update of the jail state from jls output
        """
        try:
            import json
            stdout = subprocess.check_output([
                "/usr/sbin/jls",
                "-j",
                self.identifier,
                "-v",
                "-h",
                "--libxo=json"
            ], shell=False, stderr=subprocess.DEVNULL)
            output = stdout.decode().strip()

            self.jail_state = json.loads(output)["jail-information"]["jail"][0]

        except:
            self.jail_state = None

    def _teardown_mounts(self) -> None:

        mountpoints = list(map(
            self._get_absolute_path_from_jail_asset,
            [
                "/dev/fd",
                "/dev",
                "/proc",
                "/root/compat/linux/proc"
            ]
        ))

        mountpoints += list(map(
            lambda x: x["destination"],
            list(self.fstab)
        ))

        for mountpoint in mountpoints:
            if os.path.isdir(str(mountpoint)):
                iocage.lib.helpers.umount(
                    mountpoint,
                    force=True,
                    logger=self.logger,
                    ignore_error=True  # maybe it was not mounted
                )

    def _get_absolute_path_from_jail_asset(
        self,
        value: str
    ) -> iocage.lib.Types.AbsolutePath:

        return iocage.lib.Types.AbsolutePath(f"{self.root_path}{value}")

    def _resolve_name(self, text) -> str:

        if (text is None) or (len(text) == 0):
            raise iocage.lib.errors.JailNotSupplied(logger=self.logger)

        jails_dataset = self.host.datasets.jails

        for dataset in list(jails_dataset.children):

            dataset_name = dataset.name[(len(jails_dataset.name) + 1):]
            humanreadable_name = iocage.lib.helpers.to_humanreadable_name(
                dataset_name
            )

            if text in [dataset_name, humanreadable_name]:
                return dataset_name

        raise iocage.lib.errors.JailNotFound(text, logger=self.logger)

    @property
    def name(self) -> str:
        """
        The name (formerly UUID) of the Jail
        """
        return self.config["id"]

    @property
    def humanreadable_name(self) -> str:
        """
        A human-readable identifier to print in logs and CLI output

        Whenever a Jail is found to have a UUID as identifier,
        a shortened string of the first 8 characters is returned
        """
        try:
            return iocage.lib.helpers.to_humanreadable_name(self.name)
        except:
            raise iocage.lib.errors.JailUnknownIdentifier(
                logger=self.logger
            )

    @property
    def stopped(self) -> bool:
        """
        Boolean value that is True if a jail is stopped
        """
        return self.running is not True

    @property
    def running(self) -> bool:
        """
        Boolean value that is True if a jail is running
        """
        return self.jid is not None

    @property
    def jid(self) -> int:
        """
        The JID of a running jail or None if the jail is not running
        """
        try:
            return int(self.jail_state["jid"])
        except (TypeError, AttributeError, KeyError):
            pass

        try:
            self.update_jail_state()
            return int(self.jail_state["jid"])
        except (TypeError, AttributeError, KeyError):
            return None

    @property
    def env(self):
        """
        Environment variables for hook scripts
        """
        jail_env = os.environ.copy()

        for prop in self.config.all_properties:
            prop_name = f"IOCAGE_{prop.upper()}"
            jail_env[prop_name] = self.getstring(prop)

        jail_env["IOCAGE_JAIL_PATH"] = self.root_dataset.mountpoint

        return jail_env

    @property
    def identifier(self):
        """
        Used internally to identify jails (in snapshots, jls, etc)
        """
        return f"ioc-{self.config['id']}"

    @property
    def release(self):
        """
        The iocage.Release instance linked with the jail
        """
        return iocage.lib.Release.Release(
            name=self.config["release"],
            logger=self.logger,
            host=self.host,
            zfs=self.zfs
        )

    @property
    def logfile_path(self):
        """
        Absolute path of the jail log file
        """
        return f"{self.host.datasets.logs.mountpoint}-console.log"

    def __getattribute__(self, key: str):

        try:
            return object.__getattribute__(self, key)
        except AttributeError:
            pass

        try:
            jail_state = object.__getattribute__(self, "jail_state")
        except:
            jail_state = None
            raise

        if jail_state is not None:
            try:
                return jail_state[key]
            except:
                pass

        raise AttributeError(f"Jail property {key} not found")

    def __dir__(self):

        properties = set()

        for prop in dict.__dir__(self):
            if not prop.startswith("_"):
                properties.add(prop)

        return list(properties)


class Jail(JailGenerator):

    def start(self, *args, **kwargs):
        return list(JailGenerator.start(self, *args, **kwargs))

    def stop(self, *args, **kwargs):
        return list(JailGenerator.stop(self, *args, **kwargs))



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import re
from typing import Any, Iterable, List, Union

import iocage.lib.errors
import iocage.lib.Resource


def match_filter(value: str, filter_string: str):
    escaped_characters = [".", "$", "^", "(", ")", "?"]
    for character in escaped_characters:
        filter_string = filter_string.replace(character, f"\\{character}")
    filter_string = filter_string.replace("*", ".*")
    filter_string = filter_string.replace("+", ".+")
    pattern = f"^{filter_string}$"
    match = re.match(pattern, value)
    return match is not None


class Term(list):

    glob_characters = ["*", "+"]

    def __init__(self, key, values=list()):
        self.key = key

        if values is None:
            raise TypeError("Values may not be empty")
        elif isinstance(values, str):
            data = self._split_filter_values(values)
        else:
            data = [values]

        list.__init__(self, data)

    def matches_resource(
        self,
        resource: 'iocage.lib.Resource.Resource'
    ) -> bool:
        value = resource.get(self.key)

        return self.matches(value)

    def matches(self, value: Any, short: bool=False) -> bool:
        """
        Returns True if the value matches the term

        Args:

            value:
                The value that is matched against the filter term

            short:
                When value has a length of 8 characters, this argument allows
                to match a jail's shortname as well. This is required for
                selecting jails with UUIDs by the first part of the name
        """

        # match any item of a list
        if (value is not None) and isinstance(value, list):
            # `short` not required here
            return any(map(self.matches, value))

        input_value = iocage.lib.helpers.to_string(value)

        for filter_value in self:

            if match_filter(input_value, filter_value):
                return True

            if short is False:
                continue

            # match against humanreadable names as well
            has_humanreadble_length = (len(filter_value) == 8)
            has_no_globs = not self._filter_string_has_globs(filter_value)
            if (has_humanreadble_length and has_no_globs) is True:
                shortname = iocage.lib.helpers.to_humanreadable_name(
                    input_value
                )
                if match_filter(shortname, filter_value):
                    return True

        return False

    def _filter_string_has_globs(self, filter_string: str) -> bool:
        for glob in self.glob_characters:
            if glob in filter_string:
                return True
        return False

    def _split_filter_values(self, user_input: str) -> List[str]:
        values: List[str] = []
        escaped_comma_blocks = map(
            lambda block: block.split(","),
            user_input.split("\\,")
        )
        for block in escaped_comma_blocks:
            n = len(values)
            if n > 0:
                index = n - 1
                values[index] += f",{block[0]}"
            else:
                values.append(block[0])
            if len(block) > 1:
                values += block[1:]
        return values

    def _validate_name_filter_string(self, filter_string: str) -> bool:

        globs = self.glob_characters

        # Allow glob only filters
        if (len(filter_string) == 1) and (filter_string in globs):
            return True

        # replace all glob charaters in user input
        filter_string_without_globs = ""
        for i, char in enumerate(filter_string):
            if char not in globs:
                filter_string_without_globs += char

        return iocage.lib.helpers.validate_name(filter_string_without_globs)


class Terms(list):
    """
    A group of filter terms.

    Each item in this group must match for a resource to pass the filter.
    This can be interpreted as logical AND
    """

    def __init__(self, terms: Iterable[Union[Term, str]]=None) -> None:

        data: List[Union[Term, str]] = []

        if terms is not None:

            for term in terms:
                if isinstance(term, str):
                    data += self._parse_term(term)
                elif isinstance(term, Term):
                    data.append(term)

        list.__init__(self, data)

    def match_resource(
        self,
        resource: 'iocage.lib.Resource.Resource'
    ) -> bool:
        """
        Returns True if all Terms match the resource
        """

        for term in self:
            if term.matches_resource(resource) is False:
                return False

        return True

    def match_key(self, key: str, value: str) -> bool:
        """
        Check if a value matches for a given key

        Returns True if the given value matches all terms for the specified key
        Returns Fals if one of the terms does not match
        """
        for term in self:

            if term.key != key:
                continue

            short = (key == "name")

            if term.matches(value, short) is False:
                return False

        return True

    def _parse_term(self, user_input: str) -> List[Term]:

        terms = []

        try:
            prop, value = user_input.split("=", maxsplit=1)
        except:
            prop = "name"
            value = user_input

        terms.append(Term(prop, value))

        return terms



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import typing
import os.path

import iocage.lib.helpers
import iocage.lib.Config.Jail.File.Prototype

# MyPy
import iocage.lib.Resource
import iocage.lib.Jail


class FstabLine(dict):

    def __init__(self, data: dict) -> None:
        keys = data.keys()
        if "comment" not in keys:
            data["comment"] = None
        dict.__init__(self, data)

    def __str__(self) -> str:
        return _line_to_string(self)

    def __hash__(self):
        return hash(self["destination"])


class Fstab(set, iocage.lib.Config.Jail.File.Prototype.ResourceConfigFile):
    """

    Fstab configuration file wrapper

    This object allows to read, programatically edit and write fstab files.
    Bound to an iocage resource, the location of the /etc/fstab file is
    relative to the resource's root_dataset `<resource>/root`
    """
    AUTO_COMMENT_IDENTIFIER = "iocage-auto"

    release: 'iocage.lib.Release.ReleaseGenerator'
    host: 'iocage.lib.Host.HostGenerator'
    logger: 'iocage.lib.Logger.Logger'
    jail: 'iocage.lib.Jail.JailGenerator'

    def __init__(
        self,
        jail: 'iocage.lib.Jail.JailGenerator',
        release: 'iocage.lib.Release.ReleaseGenerator'=None,
        logger: 'iocage.lib.Logger.Logger'=None,
        host: 'iocage.lib.Host.HostGenerator'=None
    ) -> None:

        self.logger = iocage.lib.helpers.init_logger(self, logger)
        self.host = iocage.lib.helpers.init_host(self, host)
        self.jail = jail
        self.release = release
        set.__init__(self)

    @property
    def path(self) -> str:
        """
        Absolute fstab file path

        This is the file read from and written to.
        """
        path = f"{self.jail.dataset.mountpoint}/fstab"
        self._require_path_relative_to_resource(
            filepath=path,
            resource=self.jail
        )
        return path

    def parse_lines(
        self,
        input: str,
        ignore_auto_created: bool=True
    ) -> None:
        """
        Parses the content of a fstab file

        Args:

            input:
                The text content of an existing fstab file

            ignore_auto_created:
                Skips reading entries that were created by iocage
        """

        set.clear(self)

        for line in input.split("\n"):

            try:
                line, comment = line.split("#", maxsplit=1)
                comment = comment.strip("# ")
                ignored_comment = Fstab.AUTO_COMMENT_IDENTIFIER
                if ignore_auto_created and (comment == ignored_comment):
                    continue
            except:
                comment = None

            line = line.strip()

            if line == "":
                continue

            fragments = line.split()
            if len(fragments) != 6:
                self.logger.log(
                    f"Invalid line in fstab file {self.path}"
                    " - skipping line"
                )
                continue

            destination = os.path.abspath(fragments[1])

            new_line = FstabLine({
                "source": fragments[0],
                "destination": fragments[1],
                "type": fragments[2],
                "options": fragments[3],
                "dump": fragments[4],
                "passnum": fragments[5],
                "comment": comment
            })

            if new_line in self:
                self.logger.error(
                    "Duplicate mountpoint in fstab: "
                    f"{destination} already mounted"
                )

            self.add_line(new_line)

    def read_file(self) -> None:
        if os.path.isfile(self.path):
            with open(self.path, "r") as f:
                self._read_file_handle(f)
                self.logger.debug(f"fstab loaded from {self.path}")

    def save(self) -> None:
        with open(self.path, "w") as f:
            self._save_file_handle(f)
            self.logger.verbose(f"{self.path} written")

    def _save_file_handle(self, f) -> None:
        f.write(self.__str__())
        f.truncate()

    def _read_file_handle(self, f) -> None:
        self.parse_lines(f.read())

    def update_and_save(
        self
    ) -> None:

        if os.path.isfile(self.path):
            f = open(self.path, "r+")
            self._read_file_handle(f)
            f.seek(0)
        else:
            f = open(self.path, "w")

        self._save_file_handle(f)
        f.close()

    def update_release(
        self,
        release: 'iocage.lib.Release.ReleaseGenerator' = None
    ) -> None:
        """
        Set a new release and save the updated file
        """
        self.release = release
        self.update_and_save()

    def new_line(
        self,
        source,
        destination,
        type="nullfs",
        options="ro",
        dump="0",
        passnum="0",
        comment=None
    ) -> None:

        line = FstabLine({
            "source": source,
            "destination": destination,
            "type": type,
            "options": options,
            "dump": dump,
            "passnum": passnum,
            "comment": comment
        })

        self.add_line(line)

    def add_line(self, line: FstabLine) -> None:

        self.logger.debug(f"Adding line to fstab: {line}")
        set.add(self, line)

    @property
    def basejail_lines(self) -> typing.List[dict]:

        if self.release is None:
            return None

        if self.jail.config["basejail_type"] != "nullfs":
            return []

        basedirs = iocage.lib.helpers.get_basedir_list(
            distribution_name=self.host.distribution.name
        )

        fstab_basejail_lines = []
        release_root_path = self.release.root_dataset.mountpoint
        for basedir in basedirs:

            source = f"{release_root_path}/{basedir}"
            destination = f"{self.jail.root_dataset.mountpoint}/{basedir}"
            fstab_basejail_lines.append({
                "source": source,
                "destination": destination,
                "type": "nullfs",
                "options": "ro",
                "dump": "0",
                "passnum": "0",
                "comment": "iocage-auto"
            })

        return fstab_basejail_lines

    def __str__(self) -> str:
        return "\n".join(map(
            _line_to_string,
            list(self)
        )) + "\n"

    def __iter__(self):
        fstab_lines = list(set.__iter__(self))
        basejail_lines = self.basejail_lines
        if basejail_lines is not None:
            fstab_lines += self.basejail_lines
        return iter(fstab_lines)

    def __contains__(self, value: typing.Any) -> bool:
        for entry in self:
            if value["destination"] == entry["destination"]:
                return True
            else:
                return False
        return False


def _line_to_string(line: FstabLine) -> str:
    output = "\t".join([
        line["source"],
        line["destination"],
        line["type"],
        line["options"],
        line["dump"],
        line["passnum"]
    ])

    if line["comment"] is not None:
        comment = line["comment"]
        output += f" # {comment}"

    return output



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import os
import platform
import re

import libzfs

import iocage.lib.Datasets
import iocage.lib.DevfsRules
import iocage.lib.Distribution
import iocage.lib.Resource
import iocage.lib.helpers

# MyPy
import iocage.lib.Config.Jail.BaseConfig  # noqa: F401


class HostGenerator:

    _class_distribution = iocage.lib.Distribution.DistributionGenerator

    _devfs: 'iocage.lib.DevfsRules.DevfsRules'
    _defaults: 'iocage.lib.Resource.DefaultResource'
    releases_dataset: libzfs.ZFSDataset

    def __init__(
        self,
        root_dataset: libzfs.ZFSDataset=None,
        defaults: 'iocage.lib.Resource.DefaultResource'=None,
        zfs: 'iocage.lib.ZFS.ZFS'=None,
        logger: 'iocage.lib.Logger.Logger'=None
    ) -> None:

        self.logger = iocage.lib.helpers.init_logger(self, logger)
        self.zfs = iocage.lib.helpers.init_zfs(self, zfs)

        self.datasets = iocage.lib.Datasets.Datasets(
            root=root_dataset,
            logger=self.logger,
            zfs=self.zfs
        )
        self.distribution = self._class_distribution(
            host=self,
            logger=self.logger
        )
        if defaults is not None:
            self._defaults = defaults

    @property
    def defaults(self) -> 'iocage.lib.Resource.DefaultResource':
        if "_defaults" not in dir(self):
            self._defaults = self._load_defaults()
        return self._defaults

    @property
    def default_config(
        self
    ) -> 'iocage.lib.Config.Jail.BaseConfig.BaseConfig':
        return self.defaults.config

    def _load_defaults(self) -> 'iocage.lib.Resource.DefaultResource':
        defaults_resource = iocage.lib.Resource.DefaultResource(
            dataset=self.datasets.root,
            logger=self.logger,
            zfs=self.zfs
        )
        defaults_resource.config.read(data=defaults_resource.read_config())
        return defaults_resource

    @property
    def devfs(self) -> 'iocage.lib.DevfsRules.DevfsRules':
        """
        Lazy-loaded DevfsRules instance
        """
        if "_devfs" not in dir(self):
            self._devfs = iocage.lib.DevfsRules.DevfsRules(
                logger=self.logger
            )
        return self._devfs

    @property
    def userland_version(self) -> float:
        return float(self.release_version.partition("-")[0])

    @property
    def release_version(self):

        if self.distribution.name == "FreeBSD":
            release_version_string = os.uname()[2]
            release_version_fragments = release_version_string.split("-")

            if len(release_version_fragments) > 1:
                return "-".join(release_version_fragments[0:2])

        elif self.distribution.name == "HardenedBSD":
            pattern = re.compile(
                r"""\(hardened\/
                    (?P<release>[A-z0-9]+(?:[A-z0-9\-]+[A-z0-9]))
                    \/
                    (?P<branch>[A-z0-9]+)
                    \):""", re.X)

            return re.search(pattern, os.uname()[3])["release"].upper()

    @property
    def processor(self) -> str:
        return platform.processor()


class Host(HostGenerator):

    _class_distribution = iocage.lib.Distribution.Distribution



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import os
import platform
import re
import urllib.request
from typing import List

import requests

import iocage.lib.errors
import iocage.lib.helpers


class DistributionGenerator:

    release_name_blacklist = [
        "",
        ".",
        "..",
        "ISO-IMAGES"
    ]

    mirror_link_pattern = r"a href=\"([A-z0-9\-_\.]+)/\""

    def __init__(self, host, zfs=None, logger=None):
        self.logger = iocage.lib.helpers.init_logger(self, logger)
        self.zfs = iocage.lib.helpers.init_zfs(self, zfs)
        self.host = iocage.lib.helpers.init_host(self, host)

        self.available_releases = None

    @property
    def _class_release(self):
        return iocage.lib.Release.ReleaseGenerator

    @property
    def name(self):
        if os.uname()[2].endswith("-HBSD"):
            return "HardenedBSD"
        else:
            return platform.system()

    @property
    def mirror_url(self):

        distribution = self.name
        processor = self.host.processor

        if distribution == "FreeBSD":
            release_path = f"/pub/FreeBSD/releases/{processor}/{processor}"
            return f"http://ftp.freebsd.org{release_path}"
        elif distribution == "HardenedBSD":
            return f"http://jenkins.hardenedbsd.org/builds"
        else:
            raise iocage.lib.errors.DistributionUnknown(distribution)

    @property
    def hash_file(self):
        if self.name == "FreeBSD":
            return "MANIFEST"
        elif self.name == "HardenedBSD":
            return "CHECKSUMS.SHA256"

    def fetch_releases(self):

        self.logger.spam(f"Fetching release list from '{self.mirror_url}'")

        resource = urllib.request.urlopen(self.mirror_url)
        charset = resource.headers.get_content_charset()
        response = resource.read().decode(charset if charset else "UTF-8")

        found_releases = self._parse_links(response)
        eol_list = self._get_eol_list()

        available_releases = sorted(
            map(  # map long HardenedBSD release names
                self._map_available_release,
                filter(  # filter out other CPU architectures on HardenedBSD
                    self._filter_available_releases,
                    found_releases
                )
            )
        )

        self.available_releases = list(map(
            lambda x: self._class_release(
                name=x,
                host=self.host,
                zfs=self.zfs,
                logger=self.logger,
                eol=self._check_eol(x, eol_list)
            ),
            available_releases
        ))
        return self.available_releases

    def _map_available_release(self, release_name):
        if self.name == "HardenedBSD":
            # e.g. HardenedBSD-11-STABLE-libressl-amd64-LATEST
            return "-".join(release_name.split("-")[1:-2])
        return release_name

    def _filter_available_releases(self, release_name):
        if self.name != "HardenedBSD":
            return True
        arch = release_name.split("-")[-2:][0]
        return self.host.processor == arch

    def _get_eol_list(self) -> List[str]:
        """Scrapes the FreeBSD website and returns a list of EOL RELEASES"""
        _eol = "https://www.freebsd.org/security/unsupported.html"
        req = requests.get(_eol)
        status = req.status_code == requests.codes.ok
        eol_releases: List[str] = []
        if not status:
            req.raise_for_status()

        for eol in req.content.decode("iso-8859-1").split():
            eol_lines = eol.strip("href=").strip("/").split(">")
            # We want a dynamic EOL
            try:
                if "-RELEASE" in eol_lines[1]:
                    eol_candidate = eol_lines[1].strip('</td')
                    if eol_candidate not in eol_releases:
                        eol_releases.append(eol_candidate)
            except IndexError:
                pass

        return eol_releases

    def _check_eol(self, release: str, eol: List[str]) -> bool:
        if self.host.distribution.name == "FreeBSD":
            if release in eol:
                return True

        return False

    def get_release_trunk_file_url(self, release, filename):

        if self.host.distribution.name == "HardenedBSD":

            return "/".join([
                "https://raw.githubusercontent.com/HardenedBSD/hardenedBSD",
                release.hbds_release_branch,
                filename
            ])

        elif self.host.distribution.name == "FreeBSD":

            if release.name == "11.0-RELEASE":
                release_name = "11.0.1"
            else:
                fragments = release.name.split("-", maxsplit=1)
                release_name = f"{fragments[0]}.0"

            base_url = "https://svnweb.freebsd.org/base/release"
            return f"{base_url}/{release_name}/{filename}?view=co"

    @property
    def releases(self):
        if self.available_releases is None:
            self.fetch_releases()
        return self.available_releases

    def _parse_links(self, text):
        blacklisted_releases = Distribution.release_name_blacklist
        matches = filter(lambda y: y not in blacklisted_releases,
                         map(lambda z: z.strip("\"/"),
                             re.findall(
                                 Distribution.mirror_link_pattern,
                                 text,
                                 re.MULTILINE)
                             )
                         )

        return matches


class Distribution(DistributionGenerator):

    @property
    def _class_release(self):
        return iocage.lib.Release.Release



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import typing

import iocage.lib.Config.Jail.BaseConfig


class JailConfigDefaults(iocage.lib.Config.Jail.BaseConfig.BaseConfig):

    user_properties: set = set()

    data: dict = {
        "id": None,
        "release": None,
        "boot": False,
        "priority": 0,
        "legacy": False,
        "priority": 0,
        "basejail": False,
        "defaultrouter": None,
        "defaultrouter6": None,
        "mac_prefix": "02ff60",
        "vnet": False,
        "interfaces": [],
        "ip4": "new",
        "ip4_saddrsel": 1,
        "ip4_addr": None,
        "ip6": "new",
        "ip6_saddrsel": 1,
        "ip6_addr": None,
        "resolver": "/etc/resolv.conf",
        "host_domainname": "none",
        "devfs_ruleset": 4,
        "enforce_statfs": 2,
        "children_max": 0,
        "allow_set_hostname": 1,
        "allow_sysvipc": 0,
        "allow_raw_sockets": 0,
        "allow_chflags": 0,
        "allow_mount": 0,
        "allow_mount_devfs": 0,
        "allow_mount_nullfs": 0,
        "allow_mount_procfs": 0,
        "allow_mount_zfs": 0,
        "allow_mount_tmpfs": 0,
        "allow_quotas": 0,
        "allow_socket_af": 0,
        "sysvmsg": "new",
        "sysvsem": "new",
        "sysvshm": "new",
        "exec_clean": 1,
        "exec_fib": 1,
        "exec_prestart": "/usr/bin/true",
        "exec_start": "/bin/sh /etc/rc",
        "exec_poststart": "/usr/bin/true",
        "exec_prestop": "/usr/bin/true",
        "exec_stop": "/bin/sh /etc/rc.shutdown",
        "exec_poststop": "/usr/bin/true",
        "exec_timeout": "60",
        "stop_timeout": "30",
        "mount_devfs": "1",
        "mount_fdescfs": "1",
        "securelevel": "2",
        "tags": [],
        "jail_zfs": False
    }

    def clear(self):
        dict.clear(self)
        dict.__init__(self, JailConfigDefaults.DEFAULTS)

    def __setitem__(
        self,
        key: str,
        value: typing.Any,
        **kwargs
    ):

        out = super().__setitem__(key, value, **kwargs)
        self.user_properties.add(key)
        return out

    @property
    def user_data(self) -> dict:
        data = {}
        for prop in self.user_properties:
            data[prop] = self.data[prop]
        return data



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import os.path
import sys
import fastentrypoints
from setuptools import find_packages, setup
from pip.req import parse_requirements

reqs = list(parse_requirements("requirements.txt", session="iocage"))
install_requires = list(map(lambda x: f"{x.name}{x.specifier}", reqs))
dependency_links = list(map(
    lambda x: str(x.link), 
    filter(lambda x: x.link, reqs)
))

if os.path.isdir("/usr/local/etc/init.d"):
    _data = [('/usr/local/etc/init.d', ['rc.d/ioc'])]
else:
    _data = [('/usr/local/etc/rc.d', ['rc.d/ioc'])]

if sys.version_info < (3, 6):
    exit("Only Python 3.6 and higher is supported.")

setup(
    name='iocage',
    license='BSD',
    version='0.2.12',
    description='A Python library to manage jails with iocage',
    keywords='FreeBSD jail iocage',
    author='iocage Contributors',
    author_email='authors@iocage.io',
    url='https://github.com/iocage/libiocage',
    python_requires='>=3.6',
    packages=find_packages(),
    package_data={'': ['fastentrypoints.py']},
    include_package_data=True,
    install_requires=install_requires,
    dependency_links=dependency_links,
    setup_requires=['pytest-runner'],
    entry_points={
        'console_scripts': [
            'ioc=iocage.cli:cli'
        ]
    },
    data_files=_data,
    tests_require=['pytest', 'pytest-cov', 'pytest-pep8']
)



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import typing

# MyPy
import libzfs  # noqa: F401
import iocage.lib.events  # noqa: F401
import iocage.lib.Jail  # noqa: F401
import iocage.lib.Types  # noqa: F401


class IocageException(Exception):

    def __init__(
        self,
        message: str,
        logger: 'iocage.lib.Logger.Logger'=None,
        level: str="error",
        silent: bool=False,
        append_warning: bool=False,
        warning: typing.Optional[str]=None
    ) -> None:

        if (logger is not None) and (silent is False):
            logger.__getattribute__(level)(message)
            if (append_warning is True) and (warning is not None):
                logger.warn(warning)
        else:
            super().__init__(message)


# Jails


class JailDoesNotExist(IocageException):

    def __init__(
        self,
        jail: 'iocage.lib.Jail.JailGenerator',
        *args,
        **kwargs
    ) -> None:

        msg = f"Jail '{jail.humanreadable_name}' does not exist"
        IocageException.__init__(self, msg, *args, **kwargs)


class JailAlreadyExists(IocageException):

    def __init__(
        self,
        jail: 'iocage.lib.Jail.JailGenerator',
        *args,
        **kwargs
    ) -> None:

        msg = f"Jail '{jail.humanreadable_name}' already exists"
        IocageException.__init__(self, msg, *args, **kwargs)


class JailNotRunning(IocageException):

    def __init__(
        self,
        jail: 'iocage.lib.Jail.JailGenerator',
        *args,
        **kwargs
    ) -> None:

        msg = f"Jail '{jail.humanreadable_name}' is not running"
        IocageException.__init__(self, msg, *args, **kwargs)


class JailAlreadyRunning(IocageException):

    def __init__(
        self,
        jail: 'iocage.lib.Jail.JailGenerator',
        *args,
        **kwargs
    ) -> None:

        msg = f"Jail '{jail.humanreadable_name}' is already running"
        IocageException.__init__(self, msg, *args, **kwargs)


class JailNotFound(IocageException):

    def __init__(
        self,
        text: str,
        *args,
        **kwargs
    ) -> None:

        msg = f"No jail matching '{text}' was found"
        IocageException.__init__(self, msg, *args, **kwargs)


class JailNotSupplied(IocageException):

    def __init__(self, *args, **kwargs) -> None:
        msg = f"Please supply a jail"
        IocageException.__init__(self, msg, *args, **kwargs)


class JailUnknownIdentifier(IocageException):

    def __init__(self, *args, **kwargs) -> None:
        msg = "The jail has no identifier yet"
        IocageException.__init__(self, msg, *args, **kwargs)


# Security


class SecurityViolation(IocageException):

    def __init__(self, reason, *args, **kwargs) -> None:
        msg = f"Security violation: {reason}"
        IocageException.__init__(self, msg, *args, **kwargs)


class SecurityViolationConfigJailEscape(IocageException):

    def __init__(self, file, *args, **kwargs) -> None:
        msg = f"The file {file} references a file outsite of the jail resource"
        IocageException.__init__(self, msg, *args, **kwargs)


# JailConfig


class JailConfigError(IocageException):
    pass


class InvalidJailName(JailConfigError):

    def __init__(self, *args, **kwargs) -> None:
        msg = (
            "Invalid jail name: "
            "Names have to begin and end with an alphanumeric character"
        )
        super().__init__(msg, *args, **kwargs)


class JailConigZFSIsNotAllowed(JailConfigError):

    def __init__(self, *args, **kwargs) -> None:
        msg = (
            "jail_zfs is disabled"
            "despite jail_zfs_dataset is configured"
        )
        super().__init__(msg, *args, **kwargs)


class InvalidJailConfigValue(JailConfigError):

    def __init__(
        self,
        property_name: str,
        jail: 'iocage.lib.Jail.JailGenerator'=None,
        reason: str=None,
        **kwargs
    ) -> None:

        msg = f"Invalid value for property '{property_name}'"
        if jail is not None:
            msg += f" of jail {jail.humanreadable_name}"
        if reason is not None:
            msg += f": {reason}"
        super().__init__(msg, **kwargs)


class InvalidJailConfigAddress(InvalidJailConfigValue):

    def __init__(self, value: str, **kwargs) -> None:
        reason = f"expected \"<nic>|<address>\" but got \"{value}\""
        super().__init__(
            reason=reason,
            **kwargs
        )


class JailConfigNotFound(IocageException):

    def __init__(self, config_type: str, *args, **kwargs) -> None:
        msg = f"Could not read {config_type} config"
        # This is a silent error internally used
        IocageException.__init__(self, msg, *args, **kwargs)


class DefaultConfigNotFound(IocageException, FileNotFoundError):

    def __init__(self, config_file_path: str, *args, **kwargs) -> None:
        msg = f"Default configuration not found at {config_file_path}"
        IocageException.__init__(self, msg, *args, **kwargs)


# General


class IocageNotActivated(IocageException):

    def __init__(self, *args, **kwargs) -> None:
        msg = (
            "iocage is not activated yet - "
            "please run `iocage activate` first and select a pool"
        )
        super().__init__(msg, *args, **kwargs)


class MustBeRoot(IocageException):

    def __init__(self, msg: str, *args, **kwargs) -> None:
        _msg = (
            f"Must be root to {msg}"
        )
        super().__init__(_msg, *args, **kwargs)


class CommandFailure(IocageException):

    def __init__(self, returncode: int, *args, **kwargs) -> None:
        msg = f"Command exited with {returncode}"
        super().__init__(msg, *args, **kwargs)


class NotAnIocageZFSProperty(IocageException):

    def __init__(self, property_name: str, *args, **kwargs) -> None:
        msg = f"The ZFS property '{property_name}' is not managed by iocage"
        super().__init__(msg, *args, **kwargs)


# Host, Distribution


class DistributionUnknown(IocageException):

    def __init__(self, distribution_name: str, *args, **kwargs) -> None:
        msg = f"Unknown Distribution: {distribution_name}"
        super().__init__(msg, *args, **kwargs)


# Storage


class UnmountFailed(IocageException):

    def __init__(
        self,
        mountpoint: iocage.lib.Types.AbsolutePath,
        *args,
        **kwargs
    ) -> None:

        msg = f"Failed to unmount {mountpoint}"
        super().__init__(msg, *args, **kwargs)


class MountFailed(IocageException):

    def __init__(
        self,
        mountpoint: iocage.lib.Types.AbsolutePath,
        *args,
        **kwargs
    ) -> None:

        msg = f"Failed to mount {mountpoint}"
        super().__init__(msg, *args, **kwargs)


class DatasetNotMounted(IocageException):

    def __init__(
        self,
        dataset: 'libzfs.ZFSDataset',
        *args,
        **kwargs
    ) -> None:

        msg = f"Dataset '{dataset.name}' is not mounted"
        super().__init__(msg, *args, **kwargs)


class DatasetNotAvailable(IocageException):

    def __init__(self, dataset_name, *args, **kwargs) -> None:
        msg = f"Dataset '{dataset_name}' is not available"
        super().__init__(msg, *args, **kwargs)


class DatasetNotJailed(IocageException):

    def __init__(
        self,
        dataset: 'libzfs.ZFSDataset',
        *args,
        **kwargs
    ) -> None:

        name = dataset.name
        msg = f"Dataset {name} is not jailed."
        warning = f"Run 'zfs set jailed=on {name}' to allow mounting"
        kwargs["append_warning"] = warning
        super().__init__(msg, *args, **kwargs)


class ZFSPoolInvalid(IocageException, TypeError):

    def __init__(
        self,
        consequence: str,
        *args,
        **kwargs
    ) -> None:

        msg = "Invalid ZFS pool"

        if consequence is not None:
            msg += f": {consequence}"

        IocageException.__init__(self, msg, *args, **kwargs)


class ZFSPoolUnavailable(IocageException):

    def __init__(
        self,
        pool_name: str,
        *args,
        **kwargs
    ) -> None:

        msg = f"ZFS pool '{pool_name}' is UNAVAIL"
        super().__init__(msg, *args, **kwargs)


# Network


class VnetBridgeMissing(IocageException):

    def __init__(self, *args, **kwargs) -> None:
        msg = "VNET is enabled and requires setting a bridge"
        super().__init__(msg, *args, **kwargs)


class InvalidNetworkBridge(IocageException, ValueError):

    def __init__(
        self,
        reason: str=None,
        *args,
        **kwargs
    ) -> None:

        msg = "Invalid network bridge argument"
        if reason is not None:
            msg += f": {reason}"
        super().__init__(msg, *args, **kwargs)


# Release


class ReleaseUpdateFailure(IocageException):

    def __init__(
        self,
        release_name: str,
        reason: str=None,
        *args,
        **kwargs
    ) -> None:

        msg = f"Release update of '{release_name}' failed"
        if reason is not None:
            msg += f": {reason}"
        super().__init__(msg, *args, **kwargs)


class InvalidReleaseAssetSignature(ReleaseUpdateFailure):

    def __init__(
        self,
        release_name: str,
        asset_name: str,
        *args,
        **kwargs
    ) -> None:

        msg = f"Asset {asset_name} has an invalid signature"
        ReleaseUpdateFailure.__init__(
            self,
            release_name=release_name,
            reason=msg,
            **kwargs
        )


class IllegalReleaseAssetContent(ReleaseUpdateFailure):

    def __init__(
        self,
        release_name: str,
        asset_name: str,
        reason: str,
        *args,
        **kwargs
    ) -> None:

        msg = f"Asset {asset_name} contains illegal files - {reason}"
        ReleaseUpdateFailure.__init__(
            self,
            release_name=release_name,
            reason=msg,
            **kwargs
        )


class ReleaseNotFetched(IocageException):

    def __init__(self, name: str, *args, **kwargs) -> None:
        msg = f"Release '{name}' does not exist or is not fetched locally"
        super().__init__(msg, *args, **kwargs)


class ReleaseUpdateBranchLookup(IocageException):

    def __init__(
        self,
        release_name: str,
        reason: str=None,
        *args,
        **kwargs
    ) -> None:

        msg = f"Update source of release '{release_name}' not found"
        if reason is not None:
            msg += ": {reason}"
        super().__init__(msg, *args, **kwargs)


# Prompts


class DefaultReleaseNotFound(IocageException):

    def __init__(self, host_release_name: str, *args, **kwargs) -> None:
        msg = (
            f"Release '{host_release_name}' not found: "
            "Could not determine a default source"
        )
        super().__init__(msg, *args, **kwargs)


# DevfsRules


class DevfsRuleException(IocageException):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)


class InvalidDevfsRulesSyntax(DevfsRuleException):

    def __init__(
        self,
        devfs_rules_file: str,
        reason: str=None,
        *args,
        **kwargs
    ) -> None:

        msg = f"Invalid devfs rules in {devfs_rules_file}"
        if reason is not None:
            msg += f": {reason}"
        super().__init__(msg, *args, **kwargs)


class DuplicateDevfsRuleset(DevfsRuleException):

    def __init__(
        self,
        devfs_rules_file: str,
        reason: str=None,
        *args,
        **kwargs
    ) -> None:

        msg = "Cannot add duplicate ruleset"
        if reason is not None:
            msg += f": {reason}"
        super().__init__(msg, *args, **kwargs)


# Logger


class LogException(IocageException):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)


class CannotRedrawLine(LogException):

    def __init__(self, reason: str, *args, **kwargs) -> None:
        msg = "Logger can't redraw line"
        if reason is not None:
            msg += f": {reason}"
        super().__init__(msg, *args, **kwargs)


# Events


class EventAlreadyFinished(IocageException):

    def __init__(
        self,
        event: 'iocage.lib.events.IocageEvent',
        *args,
        **kwargs
    ) -> None:

        msg = "This {event.type} event is already finished"
        IocageException.__init__(self, msg, *args, **kwargs)


# Jail Filter


class JailFilterException(IocageException):

    def __init__(self, *args, **kwargs) -> None:
        IocageException.__init__(self, *args, **kwargs)


class JailFilterInvalidName(JailFilterException):

    def __init__(self, *args, **kwargs) -> None:
        msg = (
            "Invalid jail selector: "
            "Cannot select jail with illegal name"
        )
        JailFilterException.__init__(self, msg, *args, **kwargs)


# Missing Features


class MissingFeature(IocageException, NotImplementedError):

    def __init__(
            self,
            feature_name: str,
            plural: bool=False,
            *args,
            **kwargs
    ) -> None:
        message = (
            f"Missing Feature: '{feature_name}' "
            "are" if plural is True else "is"
            " not implemented yet"
        )
        IocageException.__init__(self, message, *args, **kwargs)



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import typing

import iocage.lib.Config.Jail.Property

init_property = iocage.lib.Config.Jail.Property.init_property


class JailConfigProperties(dict):

    def __init__(
        self,
        config: 'iocage.lib.Config.Jail.BaseConfig.BaseConfig',
        logger: 'iocage.lib.Logger.Logger'
    ) -> None:

        self.logger = logger
        self.config = config

    def is_special_property(self, property_name: str) -> bool:
        classes = iocage.lib.Config.Jail.Property.CLASSES
        return (property_name in classes.keys())

    def get_or_create(
        self,
        property_name: str
    ) -> typing.Any:

        if property_name not in self.keys():
            self[property_name] = init_property(
                property_name=property_name,
                config=self.config,
                logger=self.logger
            )

        return self[property_name]



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import typing
from timeit import default_timer as timer

import iocage.lib.errors

# MyPy
import iocage.lib.Jail  # noqa: F401
import iocage.lib.Release  # noqa: F401

EVENT_STATUS = (
    "pending",
    "done",
    "failed"
)


class IocageEvent:
    """
    IocageEvent

    Base class for all other iocage events
    """

    HISTORY: typing.List['IocageEvent'] = []

    PENDING_COUNT: int = 0

    identifier: str = None
    _started_at: float = None
    _stopped_at: float = None
    _pending: bool = False
    skipped: bool = False
    done: bool = True
    error: BaseException = None

    def __init__(self, message=None, **kwargs) -> None:
        """
        Initializes an IocageEvent
        """

        for event in IocageEvent.HISTORY:
            if event.__hash__() == self.__hash__():
                return event  # type: ignore

        self.data = kwargs
        self.number = len(IocageEvent.HISTORY) + 1
        self.parent_count = IocageEvent.PENDING_COUNT

        self.message = message

        if self not in IocageEvent.HISTORY:
            IocageEvent.HISTORY.append(self)

    def get_state_string(
        self,
        error: str="failed",
        skipped: str="skipped",
        done: str="done",
        pending: str="pending"
    ) -> str:

        if self.error is not None:
            return error

        if self.skipped is True:
            return skipped

        if self.done is True:
            return done

        return pending

    @property
    def type(self) -> str:
        """
        The type of event

        The event type is obtained from an IocageEvent's class name
        """
        return type(self).__name__

    @property
    def pending(self) -> bool:
        return self._pending

    @pending.setter
    def pending(self, state: bool) -> None:
        current = self._pending
        new_state = (state is True)

        if current == new_state:
            return

        if new_state is True:
            if self._started_at is not None:
                raise iocage.lib.errors.EventAlreadyFinished(event=self)
            self._started_at = float(timer())
        if new_state is False:
            self._stopped_at = float(timer())

        self._pending = new_state
        IocageEvent.PENDING_COUNT += 1 if (state is True) else -1

    @property
    def duration(self) -> typing.Optional[float]:
        if (self._started_at is None) or (self._stopped_at is None):
            return None
        return self._stopped_at - self._started_at

    def _update_message(self, **kwargs) -> None:
        if "message" in kwargs:
            self.message = kwargs["message"]

    def begin(self, **kwargs) -> 'IocageEvent':
        self._update_message(**kwargs)
        self.pending = True
        self.done = False
        self.parent_count = IocageEvent.PENDING_COUNT - 1
        return self

    def end(self, **kwargs) -> 'IocageEvent':
        self._update_message(**kwargs)
        self.done = True
        self.pending = False
        self.done = True
        self.parent_count = IocageEvent.PENDING_COUNT
        return self

    def step(self, **kwargs) -> 'IocageEvent':
        self._update_message(**kwargs)
        self.parent_count = IocageEvent.PENDING_COUNT
        return self

    def skip(self, **kwargs) -> 'IocageEvent':
        self._update_message(**kwargs)
        self.skipped = True
        self.pending = False
        self.parent_count = IocageEvent.PENDING_COUNT
        return self

    def fail(self, exception=True, **kwargs) -> 'IocageEvent':
        self._update_message(**kwargs)
        self.error = exception
        self.pending = False
        self.parent_count = IocageEvent.PENDING_COUNT
        return self

    def __hash__(self):
        identifier = "generic" if self.identifier is None else self.identifier
        return hash((self.type, identifier))


# Jail


class JailEvent(IocageEvent):

    def __init__(
        self,
        jail: 'iocage.lib.Jail.JailGenerator',
        **kwargs
    ) -> None:

        try:
            self.identifier = jail.humanreadable_name
        except:
            self.identifier = None

        IocageEvent.__init__(self, jail=jail, **kwargs)


class JailLaunch(JailEvent):

    def __init__(
        self,
        jail: 'iocage.lib.Jail.JailGenerator',
        **kwargs
    ) -> None:

        JailEvent.__init__(self, jail, **kwargs)


class JailVnetConfiguration(JailEvent):

    def __init__(
        self,
        jail: 'iocage.lib.Jail.JailGenerator',
        **kwargs
    ) -> None:

        JailEvent.__init__(self, jail, **kwargs)


class JailZfsShareMount(JailEvent):

    def __init__(
        self,
        jail: 'iocage.lib.Jail.JailGenerator',
        **kwargs
    ) -> None:

        JailEvent.__init__(self, jail, **kwargs)


class JailServicesStart(JailEvent):

    def __init__(
        self,
        jail: 'iocage.lib.Jail.JailGenerator',
        **kwargs
    ) -> None:

        JailEvent.__init__(self, jail, **kwargs)


class JailDestroy(JailEvent):

    def __init__(
        self,
        jail: 'iocage.lib.Jail.JailGenerator',
        **kwargs
    ) -> None:

        JailEvent.__init__(self, jail, **kwargs)


class JailNetworkTeardown(JailEvent):

    def __init__(
        self,
        jail: 'iocage.lib.Jail.JailGenerator',
        **kwargs
    ) -> None:

        JailEvent.__init__(self, jail, **kwargs)


class JailMountTeardown(JailEvent):

    def __init__(
        self,
        jail: 'iocage.lib.Jail.JailGenerator',
        **kwargs
    ) -> None:

        JailEvent.__init__(self, jail, **kwargs)

# Release


class ReleaseEvent(IocageEvent):

    def __init__(
        self,
        release: 'iocage.lib.Release.ReleaseGenerator',
        **kwargs
    ) -> None:

        try:
            self.identifier = release.name
        except:
            self.identifier = None

        IocageEvent.__init__(self, release=release, **kwargs)


class FetchRelease(ReleaseEvent):

    def __init__(
        self,
        release: 'iocage.lib.Release.ReleaseGenerator',
        **kwargs
    ) -> None:

        ReleaseEvent.__init__(self, release, **kwargs)


class ReleasePrepareStorage(FetchRelease):

    def __init__(
        self,
        release: 'iocage.lib.Release.ReleaseGenerator',
        **kwargs
    ) -> None:

        FetchRelease.__init__(self, release, **kwargs)


class ReleaseDownload(FetchRelease):

    def __init__(
        self,
        release: 'iocage.lib.Release.ReleaseGenerator',
        **kwargs
    ) -> None:

        FetchRelease.__init__(self, release, **kwargs)


class ReleaseExtraction(FetchRelease):

    def __init__(
        self,
        release: 'iocage.lib.Release.ReleaseGenerator',
        **kwargs
    ) -> None:

        FetchRelease.__init__(self, release, **kwargs)


class ReleaseCopyBase(FetchRelease):

    def __init__(
        self,
        release: 'iocage.lib.Release.ReleaseGenerator',
        **kwargs
    ) -> None:

        FetchRelease.__init__(self, release, **kwargs)


class ReleaseConfiguration(FetchRelease):

    def __init__(
        self,
        release: 'iocage.lib.Release.ReleaseGenerator',
        **kwargs
    ) -> None:

        FetchRelease.__init__(self, release, **kwargs)


class ReleaseUpdate(ReleaseEvent):

    def __init__(
        self,
        release: 'iocage.lib.Release.ReleaseGenerator',
        **kwargs
    ) -> None:

        ReleaseEvent.__init__(self, release, **kwargs)


class ReleaseUpdateDownload(ReleaseUpdate):

    def __init__(
        self,
        release: 'iocage.lib.Release.ReleaseGenerator',
        **kwargs
    ) -> None:

        ReleaseUpdate.__init__(self, release, **kwargs)


class RunReleaseUpdate(ReleaseUpdate):

    def __init__(
        self,
        release: 'iocage.lib.Release.ReleaseGenerator',
        **kwargs
    ) -> None:

        ReleaseUpdate.__init__(self, release, **kwargs)


class ExecuteReleaseUpdate(ReleaseUpdate):

    def __init__(
        self,
        release: 'iocage.lib.Release.ReleaseGenerator',
        **kwargs
    ) -> None:

        ReleaseUpdate.__init__(self, release, **kwargs)



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
"""set module for the cli."""
import typing
import click

import iocage.lib.Logger
import iocage.lib.helpers
import iocage.lib.Resource
import iocage.lib.Jails

__rootcmd__ = True


@click.command(
    context_settings=dict(max_content_width=400,),
    name="set",
    help="Sets the specified property."
)
@click.pass_context
@click.argument("props", nargs=-1)
@click.argument("jail", nargs=1, required=True)
def cli(ctx, props, jail):
    """Get a list of jails and print the property."""

    logger = ctx.parent.logger
    host = iocage.lib.Host.HostGenerator(logger=logger)

    # Defaults
    if jail == "defaults":
        updated_properties = set_properties(props, host.defaults)
        if len(updated_properties) > 0:
            logger.screen("Defaults updated: " + ", ".join(updated_properties))
        else:
            logger.screen("Defaults unchanged")
        return

    # Jail Properties
    filters = (f"name={jail}",)
    ioc_jails = iocage.lib.Jails.JailsGenerator(
        filters,
        host=host,
        logger=logger
    )

    updated_jail_count = 0

    for jail in ioc_jails:

        updated_properties = set_properties(props, jail)

        if len(updated_properties) == 0:
            logger.screen(f"Jail '{jail.humanreadable_name}' unchanged")
        else:
            logger.screen(
                f"Jail '{jail.humanreadable_name}' updated: " +
                ", ".join(updated_properties)
            )

        updated_jail_count += 1

    if updated_jail_count == 0:
        logger.error("No jails to update")
        exit(1)

    exit(0)


def set_properties(
    properties: typing.List[str],
    target: 'iocage.lib.LaunchableResource.LaunchableResource'
) -> set:

    updated_properties = set()

    for prop in properties:

        if _is_setter_property(prop):
            key, value = prop.split("=", maxsplit=1)
            changed = target.config.set(key, value)
            if changed:
                updated_properties.add(key)
        else:
            key = prop
            try:
                del target.config[key]
                updated_properties.add(key)
            except:
                pass

    if len(updated_properties) > 0:
        target.save()

    return updated_properties


def _is_setter_property(property_string):
    return "=" in property_string



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import typing
import datetime
import hashlib
import os
import shutil
import tarfile
import urllib.request
import uuid
from urllib.parse import urlparse

import libzfs
import ucl

import iocage.lib.ZFS
import iocage.lib.errors
import iocage.lib.helpers
import iocage.lib.events
import iocage.lib.LaunchableResource
import iocage.lib.Jail

# MyPy
import iocage.lib.Resource


class ReleaseResource(iocage.lib.LaunchableResource.LaunchableResource):

    _release: 'ReleaseGenerator' = None

    def __init__(
        self,
        host: 'iocage.lib.Host.HostGenerator',
        release: 'ReleaseGenerator'=None,
        **kwargs
    ) -> None:

        self.__releases_dataset_name = host.datasets.releases.name
        self.__base_dataset_name = host.datasets.base.name
        self.host = iocage.lib.helpers.init_host(self, host)

        iocage.lib.LaunchableResource.LaunchableResource.__init__(
            self,
            **kwargs
        )

        self._release = release

    @property
    def release(self) -> 'ReleaseGenerator':
        """
        Release instance that belongs to the resource

        Usually the resource becomes inherited from the Release itself.
        It can still be used linked to a foreign ReleaseGenerator by passing
        release as named attribute to the __init__ function
        """
        if self._release is not None:
            return self._release

        elif isinstance(self, ReleaseGenerator):
            return self

        raise Exception(
            "Resource is not a valid release itself and has no linked release"
        )

    @property
    def dataset_name(self) -> str:
        """
        Name of the release base ZFS dataset

        If the resource has no dataset or dataset_name assigned yet,
        the release id is used to find name the dataset
        """
        try:
            return self._assigned_dataset_name
        except:
            pass

        return f"{self.__releases_dataset_name}/{self.release.name}"

    @dataset_name.setter
    def dataset_name(self, value: str):
        self._dataset_name = value

    @property
    def base_dataset(self) -> libzfs.ZFSDataset:
        # base datasets are created from releases. required to start
        # zfs-basejails
        return self.zfs.get_dataset(self.base_dataset_name)

    @property
    def base_dataset_name(self) -> str:
        return f"{self.__base_dataset_name}/{self.release.name}/root"

    @property
    def file(self) -> str:
        return None


class ReleaseGenerator(ReleaseResource):

    DEFAULT_RC_CONF_SERVICES = {
        "netif": False,
        "sendmail": False,
        "sendmail_submit": False,
        "sendmail_msp_queue": False,
        "sendmail_outbound": False
    }

    name: str = None
    eol: bool = False

    logger: 'iocage.lib.Logger.Logger'
    zfs: 'iocage.lib.ZFS.ZFS'
    host: 'iocage.lib.Host.HostGenerator'
    _resource: 'iocage.lib.Resource.Resource'
    _assets: typing.List[str]

    def __init__(
        self,
        name: str=None,
        host: 'iocage.lib.Host.HostGenerator'=None,
        zfs: 'iocage.lib.ZFS.ZFS'=None,
        logger: 'iocage.lib.Logger.Logger'=None,
        check_hashes: bool=True,
        eol: bool=False,
        **release_resource_args
    ) -> None:

        self.logger = iocage.lib.helpers.init_logger(self, logger)
        self.zfs = iocage.lib.helpers.init_zfs(self, zfs)
        self.host = iocage.lib.helpers.init_host(self, host)

        if not iocage.lib.helpers.validate_name(name):
            raise NameError(f"Invalid 'name' for Release: '{name}'")

        self.name = name
        self.eol = eol
        self._hbsd_release_branch = None

        self._hashes = None
        self.check_hashes = check_hashes is True

        ReleaseResource.__init__(
            self,
            host=self.host,
            logger=self.logger,
            zfs=self.zfs,
            **release_resource_args
        )

        self._assets = ["base"]
        if self.host.distribution.name != "HardenedBSD":
            self._assets.append("lib32")

    @property
    def resource(self) -> 'iocage.lib.Resource.Resource':
        return self._resource

    @resource.setter
    def resource(self, value: 'iocage.lib.Resource.Resource'):
        if value is None:
            self._resource = ReleaseResource(
                release=self,
                host=self.host,
                logger=self.logger,
                zfs=self.zfs
            )
        else:
            self._resource = value

    @property
    def releases_folder(self) -> str:
        return self.host.datasets.releases.mountpoint

    @property
    def download_directory(self) -> str:
        return self.dataset.mountpoint

    @property
    def root_dir(self) -> str:
        try:
            if self.root_dataset.mountpoint:
                return self.root_dataset.mountpoint
        except:
            pass

        return f"{self.releases_folder}/{self.name}/root"

    @property
    def assets(self) -> typing.List[str]:
        return self._assets

    @assets.setter
    def assets(self, value: typing.Union[typing.List[str], str]):
        value = [value] if isinstance(value, str) else value
        self._assets = list(map(
            lambda x: x if not x.endswith(".txz") else x[:-4],
            value
        ))

    @property
    def real_name(self) -> str:
        if self.host.distribution.name == "HardenedBSD":
            return f"HardenedBSD-{self.name}-{self.host.processor}-LATEST"
        return self.name

    @property
    def annotated_name(self) -> str:
        annotations = set()

        if self.eol is True:
            annotations.add("EOL")

        if self.newer_than_host is True:
            annotations.add("Newer than Host")

        if len(annotations) > 0:
            return f"{self.name} ({', ('.join(annotations)})"

        return f"{self.name}"

    @property
    def mirror_url(self) -> str:
        try:
            if self._mirror_url:
                return self._mirror_url
        except:
            pass
        return self.host.distribution.mirror_url

    @mirror_url.setter
    def mirror_url(self, value):
        url = urlparse(value)
        if url.scheme not in self._supported_url_schemes:
            raise ValueError(f"Invalid URL scheme '{url.scheme}'")
        self._mirror_url = url.geturl()

    @property
    def remote_url(self) -> str:
        return f"{self.mirror_url}/{self.real_name}"

    @property
    def available(self) -> bool:
        try:
            request = urllib.request.Request(self.remote_url, method="HEAD")
            resource = urllib.request.urlopen(request)
            return resource.getcode() == 200  # type: ignore
        except:
            return False

    @property
    def fetched(self) -> bool:
        if self.exists is False:
            return False

        root_dir_index = os.listdir(self.root_dataset.mountpoint)

        for expected_directory in ["dev", "var", "etc"]:
            if expected_directory not in root_dir_index:
                return False

        return True

    @property
    def newer_than_host(self):
        host_release_name = self._pad_release_name(self.host.release_version)
        release_name = self._pad_release_name(self.name)

        host_is_current = host_release_name.startswith("CURRENT")
        release_is_current = release_name.startswith("CURRENT")

        if release_is_current is True:
            if host_is_current is False:
                return True
            else:
                return False

        return (host_release_name < release_name)

    def _pad_release_name(self, release_name: str, digits: int=4) -> str:
        """
        pads releases with 0 until it has 4 characters before the first .
        """

        major_version_number = release_name.split("-")[0].split(".")[0]

        try:
            int(major_version_number)
            padding = str("0" * (digits-len(str(major_version_number))))
            return padding + release_name
        except:
            return release_name

    @property
    def zfs_pool(self) -> libzfs.ZFSPool:
        try:
            return self.root_dataset.pool
        except:
            return self.host.datasets.releases.pool

    @property
    def hashes(self):
        if not self._hashes:
            if not os.path.isfile(self.__get_hashfile_location()):
                self.logger.spam("hashes have not yet been downloaded")
                self._fetch_hashes()
            self._hashes = self.read_hashes()

        return self._hashes

    @property
    def _supported_url_schemes(self) -> typing.List[str]:
        return ["https", "http", "ftp"]

    @property
    def release_updates_dir(self) -> str:
        return f"{self.dataset.mountpoint}/updates"

    @property
    def hbds_release_branch(self):

        if self._hbsd_release_branch is not None:
            return self._hbsd_release_branch

        if self.fetched is False:
            raise iocage.lib.errors.ReleaseNotFetched(
                name=self.name,
                logger=self.logger
            )

        root_dataset_mountpoint = self.root_dataset.mountpoint
        source_file = f"{root_dataset_mountpoint}/etc/hbsd-update.conf"

        if not os.path.isfile(source_file):
            raise iocage.lib.errors.ReleaseUpdateBranchLookup(
                release_name=self.name,
                reason=f"{source_file} not found",
                logger=self.logger
            )

        with open(source_file, "r") as f:
            hbsd_update_conf = ucl.load(f.read())
            self._hbsd_release_branch = hbsd_update_conf["branch"]
            return self._hbsd_release_branch

    def fetch(self, update=None, fetch_updates=None):

        release_changed = False

        events = iocage.lib.events
        fetchReleaseEvent = events.FetchRelease(self)
        releasePrepareStorageEvent = events.ReleasePrepareStorage(self)
        releaseDownloadEvent = events.ReleaseDownload(self)
        releaseExtractionEvent = events.ReleaseExtraction(self)
        releaseConfigurationEvent = events.ReleaseConfiguration(self)
        releaseCopyBaseEvent = events.ReleaseCopyBase(self)

        if not self.fetched:

            yield fetchReleaseEvent.begin()
            yield releasePrepareStorageEvent.begin()

            # ToDo: allow to reach this for forced re-fetch
            self._clean_dataset()
            self.create_resource()
            self._ensure_dataset_mounted()

            yield releasePrepareStorageEvent.end()
            yield releaseDownloadEvent.begin()

            self._fetch_assets()

            yield releaseDownloadEvent.end()
            yield releaseExtractionEvent.begin()

            try:
                self._extract_assets()
            except Exception as e:
                yield releaseExtractionEvent.fail(e)
                raise

            yield releaseExtractionEvent.end()
            release_changed = True

            yield fetchReleaseEvent.end()

        else:

            yield fetchReleaseEvent.skip(
                message="already downloaded"
            )

            self.logger.verbose(
                "Release was already downloaded. Skipping download."
            )

        yield releaseConfigurationEvent.begin()
        if self._set_default_rc_conf() is True:
            yield releaseConfigurationEvent.end()
            release_changed = True
        else:
            yield releaseConfigurationEvent.skip()

        if fetch_updates is True:
            for event in ReleaseGenerator.fetch_updates(self):
                yield event

        if update is True:
            for event in ReleaseGenerator.update(self):
                if isinstance(event, iocage.lib.events.IocageEvent):
                    yield event
                else:
                    # the only non-IocageEvent is our return value
                    release_changed = event

        if release_changed:
            yield releaseCopyBaseEvent.begin()
            self._update_zfs_base()
            yield releaseCopyBaseEvent.end()
        else:
            yield releaseCopyBaseEvent.skip(message="release unchanged")

        self._cleanup()

    def _copy_to_base_release(self):
        iocage.lib.helpers.exec(
            [
                "rsync",
                "-a",
                "--delete",
                f"{self.root_dataset.mountpoint}/",
                f"{self.base_dataset.mountpoint}"
            ],
            logger=self.logger
        )

    @property
    def _base_resource(self) -> ReleaseResource:
        return ReleaseResource(
            release=self.release,
            logger=self.logger,
            host=self.host,
            zfs=self.zfs
        )
        # ToDo: Memoize ReleaseResource

    def fetch_updates(self):

        events = iocage.lib.events
        releaseUpdateDownloadEvent = events.ReleaseUpdateDownload(self)
        yield releaseUpdateDownloadEvent.begin()

        release_updates_dir = self.release_updates_dir
        release_update_download_dir = f"{release_updates_dir}"

        if os.path.isdir(release_update_download_dir):
            self.logger.verbose(
                f"Deleting existing updates in {release_update_download_dir}"
            )
            shutil.rmtree(release_update_download_dir)

        os.makedirs(release_update_download_dir)

        if self.host.distribution.name == "HardenedBSD":
            update_name = "hbsd-update"
            update_script_name = "hbsd-update"
            update_conf_name = "hbsd-update.conf"
        else:
            update_name = "freebsd-update"
            update_script_name = "freebsd-update.sh"
            update_conf_name = "freebsd-update.conf"

        files = {
            update_script_name: f"usr.sbin/{update_name}/{update_script_name}",
            update_conf_name: f"etc/{update_conf_name}",
        }

        for key in files.keys():

            remote_path = files[key]
            url = self.host.distribution.get_release_trunk_file_url(
                release=self,
                filename=remote_path
            )

            local_path = f"{release_updates_dir}/{key}"

            if os.path.isfile(local_path):
                os.remove(local_path)

            self.logger.verbose(f"Downloading {url}")
            urllib.request.urlretrieve(url, local_path)

            if key == update_script_name:
                os.chmod(local_path, 0o755)

            elif key == update_conf_name:

                if self.host.distribution.name == "FreeBSD":
                    with open(local_path, "r+") as f:
                        content = f.read()
                        f.seek(0)
                        f.write(content.replace(
                            "Components src",
                            "Components"
                        ))
                        f.truncate()

                os.chmod(local_path, 0o644)

            self.logger.debug(
                f"Update-asset {key} for release '{self.name}'"
                f" saved to {local_path}"
            )

        if self.host.distribution.name == "HardenedBSD":
            """
            Updates in Hardened BSD are directly executed in a jail, so that we
            do not need to pre-fetch the updates
            """
            self.logger.debug(
                "No pre-fetching of HardenedBSD updates required - skipping"
            )
            yield releaseUpdateDownloadEvent.skip(
                message="pre-fetching not supported on HardenedBSD"
            )

        else:
            self.logger.verbose(f"Fetching updates for release '{self.name}'")
            iocage.lib.helpers.exec([
                f"{self.release_updates_dir}/{update_script_name}",
                "-d",
                release_update_download_dir,
                "-f",
                f"{self.release_updates_dir}/{update_conf_name}",
                "--not-running-from-cron",
                "fetch"
            ], logger=self.logger)

            yield releaseUpdateDownloadEvent.end()

    def update(self):
        dataset = self.dataset
        snapshot_name = self._append_datetime(f"{dataset.name}@pre-update")

        runReleaseUpdateEvent = iocage.lib.events.RunReleaseUpdate(self)
        yield runReleaseUpdateEvent.begin()

        # create snapshot before the changes
        dataset.snapshot(snapshot_name, recursive=True)

        jail = iocage.lib.Jail.JailGenerator(
            {
                "uuid": str(uuid.uuid4()),
                "basejail": False,
                "allow_mount_nullfs": "1",
                "release": self.name,
                "securelevel": "0"
            },
            new=True,
            logger=self.logger,
            zfs=self.zfs,
            host=self.host,
            dataset=self.dataset
        )

        changed = False

        try:
            if self.host.distribution.name == "HardenedBSD":
                for event in self._update_hbsd_jail(jail):
                    if isinstance(event, iocage.lib.events.IocageEvent):
                        yield event
                    else:
                        changed = event
            else:
                for event in self._update_freebsd_jail(jail):
                    if isinstance(event, iocage.lib.events.IocageEvent):
                        yield event
                    else:
                        changed = event
            yield runReleaseUpdateEvent.end()
        except Exception as e:
            # kill the helper jail and roll back if anything went wrong
            self.logger.verbose(
                "There was an error updating the Jail - reverting the changes"
            )
            jail.stop(force=True)
            self.zfs.get_snapshot(snapshot_name).rollback(force=True)
            yield runReleaseUpdateEvent.fail(e)
            raise e

        return changed

    def snapshot(
        self,
        identifier: str,
        force: bool=False
    ) -> libzfs.ZFSSnapshot:
        """
        Returns a ZFS snapshot of the release

        Args:

            identifier:
                This string specifies the snapshots name

            force: (default=False)
                Enabling this option forces re-creation of a snapshot in case
                it already exists for the given idenfifier

        Returns:

            libzfs.ZFSSnapshot: The ZFS snapshot object found or created
        """

        snapshot_name = f"{self.dataset.name}@{identifier}"

        try:
            existing_snapshot = self.zfs.get_snapshot(snapshot_name)
            if force is False:
                self.logger.verbose(
                    f"Re-using release snapshot {self.name}@{identifier}"
                )
                return existing_snapshot
        except libzfs.ZFSException:
            existing_snapshot = None
            pass

        if existing_snapshot is not None:
            self.logger.verbose(
                f"Deleting release snapshot {self.name}@{identifier}"
            )
            existing_snapshot.delete()
            existing_snapshot = None

        self.dataset.snapshot(snapshot_name)
        return self.zfs.get_snapshot(snapshot_name)

    def _update_hbsd_jail(self, jail):

        events = iocage.lib.events
        executeReleaseUpdateEvent = events.ExecuteReleaseUpdate(self)

        for event in jail.start():
            yield event

        yield executeReleaseUpdateEvent.begin()

        update_script_path = f"{self.release_updates_dir}/hbsd-update"
        update_conf_path = f"{self.release_updates_dir}/hbsd-update.conf"

        try:

            # ToDo: as bad as print() replace passthru with a nice progress bar
            stdout = iocage.lib.helpers.exec_iter([
                update_script_path,
                "-c",
                update_conf_path,
                "-j",
                jail.identifier,
                "-V"
            ], logger=self.logger)

            for stdout_line in stdout:
                self.logger.verbose(stdout_line.strip("\n"), indent=1)

            self.logger.debug(f"Update of release '{self.name}' finished")

        except:

            raise iocage.lib.errors.ReleaseUpdateFailure(
                release_name=self.name,
                reason=(
                    "hbsd-update failed"
                ),
                logger=self.logger
            )

        yield executeReleaseUpdateEvent.end()

        for event in jail.stop():
            yield event

        self.logger.verbose(f"Release '{self.name}' updated")
        return True  # ToDo: return False if nothing was updated

    def _update_freebsd_jail(self, jail):

        events = iocage.lib.events
        executeReleaseUpdateEvent = events.ExecuteReleaseUpdate(self)

        local_update_mountpoint = f"{self.root_dir}/var/db/freebsd-update"
        if not os.path.isdir(local_update_mountpoint):
            self.logger.spam(
                "Creating mountpoint {local_update_mountpoint}"
            )
            os.makedirs(local_update_mountpoint)

        jail.fstab.new_line(
            self.release_updates_dir,
            local_update_mountpoint,
            "nullfs",
            "rw"
        )
        jail.fstab.save()

        for event in jail.start():
            yield event

        yield executeReleaseUpdateEvent.begin()
        child, stdout, stderr = jail.exec([
            "/var/db/freebsd-update/freebsd-update.sh",
            "-d",
            "/var/db/freebsd-update",
            "-f",
            "/var/db/freebsd-update/freebsd-update.conf",
            "install"
        ], ignore_error=True)

        if child.returncode != 0:
            if "No updates are available to install." in stdout:
                yield executeReleaseUpdateEvent.skip(
                    message="already up to date"
                )
                self.logger.debug("Already up to date")
            else:
                yield executeReleaseUpdateEvent.failed()
                raise iocage.lib.errors.ReleaseUpdateFailure(
                    release_name=self.name,
                    reason=(
                        "freebsd-update.sh exited "
                        f"with returncode {child.returncode}"
                    ),
                    logger=self.logger
                )
        else:
            yield executeReleaseUpdateEvent.end()
            self.logger.debug(f"Update of release '{self.name}' finished")

        for event in jail.stop():
            yield event

        self.logger.verbose(f"Release '{self.name}' updated")
        yield True  # ToDo: return False if nothing was updated

    def _append_datetime(self, text):
        now = datetime.datetime.utcnow()
        text += now.strftime("%Y%m%d%H%I%S.%f")
        return text

    def _ensure_dataset_mounted(self):
        if not self.dataset.mountpoint:
            self.dataset.mount()

    def _fetch_hashes(self):
        url = f"{self.remote_url}/{self.host.distribution.hash_file}"
        path = self.__get_hashfile_location()
        self.logger.verbose(f"Downloading hashes from {url}")
        urllib.request.urlretrieve(url, path)
        self.logger.debug(f"Hashes downloaded to {path}")

    def _fetch_assets(self):
        for asset in self.assets:
            url = f"{self.remote_url}/{asset}.txz"
            path = self._get_asset_location(asset)

            if os.path.isfile(path):
                self.logger.verbose(f"{path} already exists - skipping.")
                return
            else:
                self.logger.debug(f"Starting download of {url}")
                urllib.request.urlretrieve(url, path)
                self.logger.verbose(f"{url} was saved to {path}")

    def _clean_dataset(self):

        if not os.path.isdir(self.root_dir):
            return

        root_dir_index = os.listdir(self.root_dir)
        if not len(root_dir_index) > 0:
            return

        self.logger.verbose(
            f"Remove existing fragments from {self.root_dir}"
        )
        for directory in root_dir_index:
            asset_path = os.path.join(self.root_dir, directory)
            self.logger.spam(f"Purging {asset_path}")
            self._rmtree(asset_path)

    def read_hashes(self):
        # yes, this can read HardenedBSD and FreeBSD hash files
        path = self.__get_hashfile_location()
        hashes = {}
        with open(path, "r") as f:
            for line in f.read().split("\n"):
                s = set(line.replace("\t", " ").split(" "))
                fingerprint = None
                asset = None
                for x in s:
                    x = x.strip("()")
                    if len(x) == 64:
                        fingerprint = x
                    elif x.endswith(".txz"):
                        asset = x[:-4]
                if asset and fingerprint:
                    hashes[asset] = fingerprint
        count = len(hashes)
        self.logger.spam(f"{count} hashes read from {path}")
        return hashes

    def __get_hashfile_location(self):
        hash_file = self.host.distribution.hash_file
        return f"{self.download_directory}/{hash_file}"

    def _get_asset_location(self, asset_name):
        return f"{self.download_directory}/{asset_name}.txz"

    def _extract_assets(self):

        for asset in self.assets:

            if self.check_hashes:
                self._check_asset_hash(asset)

            with tarfile.open(self._get_asset_location(asset)) as f:

                self.logger.verbose(f"Verifying file structure in {asset}")
                self._check_tar_files(f.getmembers(), asset_name=asset)

                self.logger.debug(f"Extracting {asset}")
                f.extractall(self.root_dir)
                self.logger.verbose(
                    f"Asset {asset} was extracted to {self.root_dir}"
                )

    def _set_default_rc_conf(self) -> bool:

        for key, value in self.DEFAULT_RC_CONF_SERVICES.items():
            self.rc_conf[f"{key}_enable"] = value

        return self.rc_conf.save()

    def _generate_default_rcconf_line(self, service_name):
        if Release.DEFAULT_RC_CONF_SERVICES[service_name] is True:
            state = "YES"
        else:
            state = "NO"
        return f"{service_name}_enable=\"{state}\""

    def _update_name_from_dataset(self):
        if self.dataset is not None:
            self.name = self.dataset.name.split("/")[-2:-1]

    def _update_zfs_base(self):

        base_dataset = self.zfs.get_or_create_dataset(self.base_dataset_name)

        basedirs = iocage.lib.helpers.get_basedir_list(
            distribution_name=self.host.distribution.name
        )

        for folder in basedirs:
            self.zfs.get_or_create_dataset(f"{base_dataset.name}/{folder}")

        self._copy_to_base_release()

        self.logger.debug(f"Base release '{self.name}' updated")

    def _cleanup(self):
        for asset in self.assets:
            asset_location = self._get_asset_location(asset)
            if os.path.isfile(asset_location):
                os.remove(asset_location)

    def _check_asset_hash(self, asset_name):
        local_file_hash = self._read_asset_hash(asset_name)
        expected_hash = self.hashes[asset_name]

        has_valid_hash = local_file_hash == expected_hash
        if not has_valid_hash:
            self.logger.warn(
                f"Asset {asset_name}.txz has an invalid signature"
                f"(was '{local_file_hash}' but expected '{expected_hash}')"
            )
            raise iocage.lib.errors.InvalidReleaseAssetSignature(
                release_name=self.name,
                asset_name=asset_name,
                logger=self.logger
            )

        self.logger.spam(
            f"Asset {asset_name}.txz has a valid signature ({expected_hash})"
        )

    def _read_asset_hash(self, asset_name):
        asset_location = self._get_asset_location(asset_name)
        sha256 = hashlib.sha256()
        with open(asset_location, 'rb') as f:
            for block in iter(lambda: f.read(65536), b''):
                sha256.update(block)
        return sha256.hexdigest()

    def _check_tar_files(self, tar_infos, asset_name):
        for i in tar_infos:
            self._check_tar_info(i, asset_name)

    def _check_tar_info(self, tar_info, asset_name):
        if tar_info.name == ".":
            return
        if not tar_info.name.startswith("./"):
            reason = "Names in txz files must be relative and begin with './'"
        elif ".." in tar_info.name:
            reason = "Names in txz files must not contain '..'"
        else:
            return

        raise iocage.lib.errors.IllegalReleaseAssetContent(
            release_name=self.name,
            asset_name=asset_name,
            reason=reason,
            logger=self.logger
        )

    def __str__(self):
        return self.name

    def destroy(self):
        self.storage.delete_dataset_recursive(self.dataset)


class Release(ReleaseGenerator):

    def fetch(self, *args, **kwargs):
        return list(ReleaseGenerator.fetch(self, *args, **kwargs))

    def update(self, *args, **kwargs):
        return list(ReleaseGenerator.update(self, *args, **kwargs))



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
"""stop module for the cli."""
import click

import iocage.lib.errors
import iocage.lib.Jails
import iocage.lib.Logger

__rootcmd__ = True


@click.command(name="stop", help="Stops the specified jails or ALL.")
@click.pass_context
@click.option("--rc", default=False, is_flag=True,
              help="Will stop all jails with boot=on, in the specified"
                   " order with higher value for priority stopping first.")
@click.option("--log-level", "-d", default=None)
@click.option("--force", "-f", is_flag=True, default=False,
              help="Skip checks and enforce jail shutdown")
@click.argument("jails", nargs=-1)
def cli(ctx, rc, log_level, force, jails):
    """
    Looks for the jail supplied and passes the uuid, path and configuration
    location to stop_jail.
    """
    logger = ctx.parent.logger

    if len(jails) == 0:
        logger.error("No jail selector provided")
        exit(1)

    ioc_jails = iocage.lib.Jails.JailsGenerator(
        logger=logger,
        filters=jails
    )

    changed_jails = []
    failed_jails = []
    for jail in ioc_jails:
        try:
            ctx.parent.print_events(jail.stop(force=force))
        except iocage.lib.errors.IocageException:
            failed_jails.append(jail)
            continue

        logger.log(f"{jail.name} stopped")
        changed_jails.append(jail)

    if len(failed_jails) > 0:
        exit(1)

    if len(changed_jails) == 0:
        jails_input = " ".join(list(jails))
        logger.error(f"No jailes matches your input: {jails_input}")
        exit(1)



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import typing
import os.path
import iocage.lib.helpers


class Prototype:

    logger: 'iocage.lib.Logger.Logger' = None
    data: dict = {}

    def __init__(
        self,
        logger: 'iocage.lib.Logger.Logger'=None
    ) -> None:

        self.logger = iocage.lib.helpers.init_logger(self, logger)

    @property
    def file(self) -> str:
        raise NotImplementedError(
            "This needs to be implemented by the inheriting class"
        )

    @file.setter
    def file(self, value: str):
        raise NotImplementedError(
            "This needs to be implemented by the inheriting class"
        )

    def read(self) -> dict:
        try:
            with open(self.file, "r") as data:
                return self.map_input(data)
        except:
            return {}

    def write(self, data: dict) -> None:
        """
        Writes changes to the config file
        """
        with open(self.file, "w") as conf:
            conf.write(self.map_output(data))
            conf.truncate()

    def map_input(self, data: typing.Any) -> dict:
        return data

    def map_output(self, data: typing.Any) -> typing.Any:
        return data

    def exists(self) -> bool:
        return os.path.isfile(self.file)
